from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.apps import apps
from django.conf import settings
from django.core.urlresolvers import reverse
from django.contrib.auth.models import User
from django.contrib.sites.models import Site

from danceschool.core.models import DanceType, DanceTypeLevel, DanceRole, PricingTier, InstructorListPluginModel

from dynamic_preferences.registries import global_preferences_registry
import re
from six.moves import input
from importlib import import_module

try:
    import readline
except ImportError:
    pass


class Command(BaseCommand):
    help = 'Easy-install setup script for new dance schools'

    def boolean_input(self, question, default=None):
        '''
        Method for yes/no boolean inputs
        '''
        result = input("%s: " % question)
        if not result and default is not None:
            return default
        while len(result) < 1 or result[0].lower() not in "yn":
            result = input("Please answer yes or no: ")
        return result[0].lower() == "y"

    def pattern_input(self, question, message='Invalid entry', pattern='^[a-zA-Z0-9_ ]+$', default='',required=True):
        '''
        Method for input disallowing special characters, with optionally
        specifiable regex pattern and error message.
        '''
        result = ''
        requiredFlag = True

        while (not result and requiredFlag):
            result = input('%s: ' % question)
            if result and pattern and not re.match(pattern, result):
                self.stdout.write(self.style.ERROR(message))
                result = ''
            elif not result and default:
                # Return default for fields with default
                return default
            elif not result and required:
                # Ask again for required fields
                self.stdout.write(self.style.ERROR('Answer is required.'))
            elif not required:
                # No need to re-ask for non-required fields
                requiredFlag = False
        return result

    def float_input(self, question, message='Invalid entry', default=None, required=True):
        '''
        Method for floating point inputs with optionally specifiable error message.
        '''
        float_result = None
        requiredFlag = True

        while (float_result is None and requiredFlag):
            result = input('%s: ' % question)
            if not result and not required:
                float_result = None
                requiredFlag = False
            if not result and default:
                float_result = default

            if float_result is None and requiredFlag:
                try:
                    float_result = float(result)
                except ValueError:
                    self.stdout.write(self.style.ERROR(message))
                    float_result = None
        return float_result

    def handle(self, *args, **options):

        from cms.api import create_page, add_plugin, publish_page
        from cms.constants import VISIBILITY_ANONYMOUS, VISIBILITY_USERS
        from cms.models import Page, StaticPlaceholder

        prefs = global_preferences_registry.manager()

        # Do some sanity checks to ensure that necessary apps are listed in INSTALLED_APPS
        # before proceeding
        required_apps = [
            ('cms', 'Django CMS'),
            ('danceschool.core', 'Core danceschool app'),
            ('dynamic_preferences', 'django-dynamic-preferences'),
        ]
        for this_app in required_apps:
            if not apps.is_installed(this_app[0]):
                self.stdout.write(self.style.ERROR('ERROR: %s is not installed or listed in INSTALLED_APPS. Please install before proceeding.' % this_app[1]))
                return None

        this_user = User.objects.filter(is_superuser=True).first()
        if not this_user:
            self.stdout.write(self.style.ERROR('ERROR: Superuser has not yet been created.  Please run \'python manage.py createsuperuser\' before proceeding.'))
            return None

        if Page.objects.count() > 0:
            self.stdout.write(self.style.WARNING('WARNING: CMS pages have already been created.  Creating duplicate CMS pages may lead to unexpected results.'))

        self.stdout.write(
            """
WELCOME
-------

Welcome to the django-danceschool setup script. This script is designed to guide you
through the setup of many key parameters and initial defaults that will make your
experience simpler and faster.

All inputs of this script can be later changed using the site's admin interface, so
if you decide that you need to change something later, you should not need to run this
script again.

Please answer the following questions (defaults values are in brackets)


BASIC SETUP:
------------
            """
        )

        # Current Domain
        current_domain = self.pattern_input(
            'Enter the domain name of this installation, with no protocol or trailing slashes (e.g. \'bostonlindyhop.com\') [localhost:8000]',
            message='Invalid domain name (no whitespace or slashes allowed)',
            default='localhost:8000',
            pattern='^[^/\\ ]+$'
        )
        this_site = Site.objects.get_current()
        this_site.name = current_domain
        this_site.domain = current_domain
        this_site.save()

        # School name
        school_name = self.pattern_input(
            'Give your dance school a name [Dance School]',
            message='Invalid school name (no special characters allowed)',
            default='Dance School'
        )
        prefs['contact__businessName'] = school_name
        prefs['email__defaultEmailName'] = school_name

        # School Email
        school_email = self.pattern_input(
            'Provide a default school email address',
            message='Invalid email address.',
            pattern='^[^@]+@[^@]+\.[^@]+$',
        )
        prefs['contact__businessEmail'] = school_email
        prefs['email__defaultEmailFrom'] = school_email
        prefs['email__errorEmailFrom'] = school_email

        # School Phone and Address
        school_phone = input('Provide a phone number for the school (optional): ')
        if school_phone:
            prefs['contact__businessPhone'] = school_phone

        school_address1 = input('Enter school street address - Line 1 (optional): ')
        if school_address1:
            prefs['contact__businessAddress'] = school_address1

        school_address2 = input('Enter school street address - Line 2 (optional): ')
        if school_address2:
            prefs['contact__businessAddressLineTwo'] = school_address2

        school_city = self.pattern_input(
            'Enter school city [Boston]',
            message='Invalid city (no special characters allowed)',
            pattern='^[a-zA-Z0-9 ]+$',
            default='Boston',
        )
        prefs['contact__businessCity'] = school_city

        school_state = self.pattern_input(
            'Enter school state [MA]',
            message='Invalid state (no special characters allowed)',
            pattern='^[a-zA-Z0-9 ]+$',
            default='MA',
        )
        prefs['contact__businessState'] = school_state

        school_postal = input('Enter school ZIP/postal code [02114]: ') or '02114'
        prefs['contact__businessZip'] = school_postal

        school_country = self.pattern_input(
            'Enter two-letter school country code [US]',
            message='Invalid state (only two uppercase letters allowed)',
            pattern='^[A-Z][A-Z]$',
            default='US',
        )
        prefs['contact__businessCountryCode'] = school_country

        currency_code = self.pattern_input(
            'Enter three-letter currency code for transactions [USD]',
            message='Invalid currency code (only three uppercase letters allowed)',
            pattern='^[A-Z][A-Z][A-Z]$',
            default='USD',
        )
        prefs['general__currencyCode'] = currency_code

        currency_symbol = input('Unicode currency symbol for display [$]: ') or '$'
        prefs['general__currencySymbol'] = currency_symbol

        send_error_emails = self.boolean_input('Send error emails [Y/n]',True)
        prefs['email__enableErrorEmails'] = send_error_emails

        error_email_address = self.pattern_input(
            'Email address to send error messages to [%s]' % school_email,
            message='Invalid email address.',
            pattern='^[^@]+@[^@]+\.[^@]+$',
            default=school_email,
        )
        prefs['email__errorEmailTo'] = error_email_address

        self.stdout.write(
            """
DANCE TYPES
-----------

By default, all class series have both a dance type associated with them, and a class
level associated with that dance type.  To simplify setup, we ask you to define one dance
type, and we also encourage you to set up initial levels.

Remember, you can always rename these initial values or add/delete dance levels later.

            """
        )

        initial_dancetype = self.pattern_input(
            'Name of your initial default dance type [Lindy Hop]',
            message='Invalid dance type name (no special characters)',
            default='Lindy Hop',
        )
        initial_dancetype_object = DanceType.objects.get_or_create(name=initial_dancetype, defaults={'order': 1.0})

        use_initial_levels = self.boolean_input('Use initial levels \'Level 1\', \'Level 2\', and \'Level 3\' [Y/n]', True)
        if use_initial_levels:
            DanceTypeLevel.objects.get_or_create(name='Level 1', danceType=initial_dancetype_object[0], defaults={'order': 1.0})
            DanceTypeLevel.objects.get_or_create(name='Level 2', danceType=initial_dancetype_object[0], defaults={'order': 2.0})
            DanceTypeLevel.objects.get_or_create(name='Level 3', danceType=initial_dancetype_object[0], defaults={'order': 3.0})

        self.stdout.write(
            """

ROLES
-----

Most partnered dance schools define 'Leader' and 'Follower' roles that are used in
the registration process.  If your dance school does not use these roles, or
if you do not ask students to register for a particular role, then answer 'No' below.

            """
        )

        define_roles = self.boolean_input('Define \'Leader\' and \'Follower\' roles [Y/n]', True)
        if define_roles:
            DanceRole.objects.get_or_create(name='Leader',defaults={'pluralName': 'Leaders', 'order':1.0})
            DanceRole.objects.get_or_create(name='Follower',defaults={'pluralName': 'Followers', 'order': 2.0})
            initial_dancetype_object[0].roles = DanceRole.objects.filter(name__in=['Leader','Follower'])
            initial_dancetype_object[0].save()

        self.stdout.write(
            """
PRICING
-------

All class series are priced by pricing tiers.  A pricing tier defines the default
price for any class series or event associated with that pricing tier.  Default
prices may vary depending on whether the student registers online or at the door,
and default prices may also be different for individuals who receive a student
discount.

If you have enabled the discounts app or the vouchers app, then those discounts or
vouchers are applied relative to the baseline pricing

For example, at Boston Lindy Hop, we charge a default price of $50 for online
registration and $60 for registration at the door.  For college students, we offer
an automatic discount of $10, so prices are $40 and $50, respectively.  Then, we
use the discounts app to provide further discounts for students who register for
more than one class at a time, etc.

Remember, you can always modify, add or delete pricing tiers from the admin
interface after completing this setup.

            """
        )

        initial_pricing_tier_name = self.pattern_input(
            'Give your initial pricing tier a name [Default Pricing]',
            message='Invalid dance type name (no special characters)',
            default='Default Pricing',
        )

        initial_online_price = self.float_input('Initial online price [50]',default=50)
        initial_door_price = self.float_input('Initial at-the-door price [60]',default=60)
        initial_dropin_price = self.float_input('Initial price for single-class drop-ins [15]',default=15)

        PricingTier.objects.get_or_create(
            name=initial_pricing_tier_name,
            defaults={
                'onlinePrice': initial_online_price,
                'doorPrice': initial_door_price,
                'dropinPrice': initial_dropin_price,
            })

        if apps.is_installed('danceschool.vouchers'):
            enable_vouchers = self.boolean_input('Enable the voucher/gift certificate system [Y/n]', True)
            prefs['vouchers__enableVouchers'] = enable_vouchers
            prefs['vouchers__enableGiftCertificates'] = enable_vouchers
            prefs['vouchers__enableGiftCertificatePDF'] = enable_vouchers

        if apps.is_installed('danceschool.discounts'):
            enable_discounts = self.boolean_input('Enable the discount system [Y/n]', True)
            prefs['general__discountsEnabled'] = enable_discounts

        if apps.is_installed('danceschool.financial'):
            self.stdout.write(
                """
FINANCIAL MANAGEMENT
--------------------

This project contains a financial app which allows you to manage revenues and expenses
in an automated way. In particular, it is possible for expense items related to instruction
and venue rental to be automatically generated, and it is possible for
revenue items related to registrations to be automatically generated as well.  However,
if you do not want these features, then you may disable them below.

In order to automatic generation of expenses to be functional, you will also need
to set compensation rates for your Locations and Instructors as you create them.
Expenses may be generated on an hourly basis (for hourly rental/compensation), or on
a daily/weekly/monthly ongoing basis as well.

                """
            )

            generate_staff = self.boolean_input('Auto-generate staff expense items for completed events [Y/n]', True)
            prefs['financial__autoGenerateExpensesEventStaff'] = generate_staff

            if generate_staff:
                # This just ensures that the standard staff categories are created before launch
                prefs.get('financial__classInstructionExpenseCat',None)
                prefs.get('financial__assistantClassInstructionExpenseCat',None)
                prefs.get('financial__otherStaffExpenseCat',None)

            generate_venue = self.boolean_input('Auto-generate venue rental expense items for completed events [Y/n]', True)
            prefs['financial__autoGenerateExpensesVenueRental'] = generate_venue

            generate_registration = self.boolean_input('Auto-generate registration revenue items for registrations [Y/n]', True)
            prefs['financial__autoGenerateRevenueRegistrations'] = generate_registration

        if apps.is_installed('danceschool.private_lessons'):
            self.stdout.write(
                """
PRIVATE LESSONS
---------------

In addition to providing the ability to register for class series and public events,
this project also optionally allows for the scheduling of and registration for
private lessons with instructors.

Depending on the way in which your school is structured, you may choose to permit
or disable the general public from registering for private lessons.  Disabling
this feature may be desirable if you, for example, only allow staff to book lessons.
You may also choose to enable the full registration process online, or to disable
it if you would prefer that students pay instructors directly at the time of their lesson.
                """
            )

            allow_public_privatelesson_booking = self.boolean_input('Allow public booking of private lessons [Y/n]', True)
            prefs['privateLessons__allowPublicBooking'] = allow_public_privatelesson_booking
            allow_privatelesson_payment = self.boolean_input('Allow payment for private lessons through the registration system [Y/n]', True)
            prefs['privateLessons__allowRegistration'] = allow_privatelesson_payment
            notify_privatelesson_instructor = self.boolean_input('Notify private lesson instructors on lesson booking [Y/n]', True)
            prefs['privateLessons__notifyInstructor'] = notify_privatelesson_instructor

        self.stdout.write(
            """
INITIAL PAGES
-------------

The django-danceschool project uses Django CMS, which allows for highly configurable pages
with plugins to display complex functionality.  However, it is often helpful not to have
to start with a completely blank site.  The next few questions will allow you to automatically
set up some of the features that most dance schools use.

There are two basic methods in which pages can be configured:
    1. A traditional setup with a "Home" page and the "Registration" page located on a
       sub-page, typically "/register/"
    2. A "registration-only" setup in which the registration page is located at your
       site's base URL instead of a home page.  This option is recommended for individuals
       who wish to use the registration system while maintaining other content on an
       external website.

Remember, all page settings and content can be changed later via the admin interface.

            """
        )

        try:
            initial_language = settings.LANGUAGES[0][0]
        except IndexError:
            initial_language = getattr(settings, 'LANGUAGE_CODE', 'en')

        # First, add the footer if no footer exists
        sp = StaticPlaceholder.objects.get_or_create(code='footer')
        sp_draft = sp[0].draft
        sp_public = sp[0].public
        if not sp_public.cmsplugin_set.all():
            address_string = '{}{}{}, {} {}<br />'.format(
                school_address1 + '<br />\n' if school_address1 else '',
                school_address2 + '<br />\n' if school_address2 else '',
                school_city, school_state, school_postal
            )
            email_string = '<a href="mailto:{}">{}</a>'.format(school_email, school_email)
            initial_footer = '<p class="text-center"><strong>%s</strong><br />\n%s\n%s</p>' % (school_name, address_string, email_string)
            add_plugin(sp_draft, 'TextPlugin', initial_language, body=initial_footer)
            add_plugin(sp_public, 'TextPlugin', initial_language, body=initial_footer)

        registration_first = self.boolean_input('Perform a "registration-only" setup with registration on the home page? [y/N]', False)
        if registration_first:
            home_page = create_page(
                'Registration', 'cms/home.html', initial_language, menu_title='Registration',
                apphook='RegistrationApphook', in_navigation=True, published=True
            )
            home_page.set_as_homepage()
            self.stdout.write('Registration page added.\n')
        else:
            add_home_page = self.boolean_input('Create a \'Home\' page [Y/n]', True)
            if add_home_page:
                home_page = create_page('Home', 'cms/frontpage.html', initial_language, menu_title='Home', in_navigation=True, published=True)
                content_placeholder = home_page.placeholders.get(slot='content')
                add_plugin(content_placeholder, 'TextPlugin', initial_language, body='<h1>Welcome to %s</h1>\n\n<p>If you are logged in, click \'Edit Page\' to begin adding content.</p>' % school_name)
                publish_page(home_page, this_user, initial_language)
                home_page.set_as_homepage()
                self.stdout.write('Home page added.\n')
            add_registration_link = self.boolean_input('Add a link to the Registration page to the main navigation menu [Y/n]', True)
            if add_registration_link:
                registration_link_page = create_page(
                    'Registration', 'cms/home.html', initial_language,
                    menu_title='Register', slug='register', overwrite_url=reverse('registration'), in_navigation=True, published=True
                )
                self.stdout.write('Registration link added.\n')

        add_instructor_page = self.boolean_input('Add a page to list all instructors with their photos and bios [Y/n]', True)
        if add_instructor_page:
            instructor_page = create_page('Instructors', 'cms/twocolumn_rightsidebar.html', initial_language, menu_title='Instructors', in_navigation=True, published=True)
            content_placeholder = instructor_page.placeholders.get(slot='content')
            sidebar_placeholder = instructor_page.placeholders.get(slot='sidebar')
            add_plugin(
                content_placeholder, 'InstructorListPlugin', initial_language,
                orderChoice=InstructorListPluginModel.OrderChoices.random,
                bioRequired=True,
                template='core/instructor_list.html'
            )
            add_plugin(
                sidebar_placeholder, 'InstructorListPlugin', initial_language,
                orderChoice=InstructorListPluginModel.OrderChoices.random,
                photoRequired=True,
                template='core/instructor_image_set.html'
            )
            publish_page(instructor_page, this_user, initial_language)
            self.stdout.write('Instructor page added.\n')

        add_calendar_page = self.boolean_input('Add a page with a public calendar of classes/events [Y/n]', True)
        if add_calendar_page:
            calendar_page = create_page('Calendar', 'cms/home.html', initial_language, menu_title='Calendar', in_navigation=True, published=True)
            content_placeholder = calendar_page.placeholders.get(slot='content')
            add_plugin(content_placeholder, 'PublicCalendarPlugin', initial_language)
            publish_page(calendar_page, this_user, initial_language)
            self.stdout.write('Calendar page added.\n')

        if apps.is_installed('danceschool.private_lessons') and allow_public_privatelesson_booking:
            add_privatelesson_link = self.boolean_input('Add a link to book private lessons to the main navigation menu [Y/n]', True)
            if add_privatelesson_link:
                privatelesson_link_page = create_page(
                    'Schedule Private Lessons', 'cms/home.html', initial_language,
                    menu_title='Private Lessons', slug='private_lessons', overwrite_url=reverse('bookPrivateLesson'), in_navigation=True, published=True
                )
                self.stdout.write('Private lesson scheduling link added.\n')

        if apps.is_installed('danceschool.faq'):
            add_faq_page = self.boolean_input('Add an FAQ page and General FAQs Category [Y/n]', True)
            if add_faq_page:
                faq_models = import_module('danceschool.faq.models')
                general_cat = faq_models.FAQCategory.objects.get_or_create(name='General Questions')
                faq_page = create_page('Frequently Asked Questions', 'cms/twocolumn_rightsidebar.html', initial_language, menu_title='FAQ', in_navigation=True, published=True)
                content_placeholder = faq_page.placeholders.get(slot='content')
                sidebar_placeholder = faq_page.placeholders.get(slot='sidebar')
                add_plugin(content_placeholder, 'FAQCategoryPlugin', initial_language, category=general_cat[0])
                add_plugin(sidebar_placeholder, 'FAQTOCPlugin', initial_language)
                publish_page(faq_page, this_user, initial_language)
                self.stdout.write('FAQ page added.\n')

        if apps.is_installed('danceschool.news'):
            add_news_page = self.boolean_input('Add a News page, along with an initial Welcome news item [Y/n]', True)
            if add_news_page:
                news_models = import_module('danceschool.news.models')
                news_models.NewsItem.objects.create(title='Welcome to %s!' % school_name, content='<p>Continue to check this news feed to remain up-to-date on everything that is happening with the school.</p>')

                create_page(
                    'Latest News', 'cms/twocolumn_rightsidebar.html', initial_language,
                    menu_title='News', apphook='NewsApphook', in_navigation=True, published=True)
                self.stdout.write('News page added.\n')

        if apps.is_installed('danceschool.stats'):
            add_stats_page = self.boolean_input('Add a private school stats page and add default graphs [Y/n]', True)
            if add_stats_page:
                stats_page = create_page(
                    'School Performance Stats', 'cms/admin_home.html', initial_language,
                    menu_title='Stats', slug='stats', apphook='StatsApphook', in_navigation=False, published=False)
                sp = StaticPlaceholder.objects.get_or_create(code='stats_graphs')
                stats_placeholder = sp[0].draft
                stats_placeholder_public = sp[0].public

                template_list = [
                    'stats/schoolstats_timeseriesbymonth.html',
                    'stats/schoolstats_averagebyclasstype.html',
                    'stats/schoolstats_averagebyclasstypemonth.html',
                    'stats/schoolstats_cohortretention.html',
                    'stats/schoolstats_averagesbylocation.html',
                    'stats/schoolstats_registrationtypes.html',
                    'stats/schoolstats_bestcustomers.html',
                ]
                for template in template_list:
                    add_plugin(stats_placeholder, 'StatsGraphPlugin', initial_language, template=template)
                    add_plugin(stats_placeholder_public, 'StatsGraphPlugin', initial_language, template=template)
                publish_page(stats_page, this_user, initial_language)
                self.stdout.write('School performance stats page added.\n')

        add_login_link = self.boolean_input('Add login/logout and account links to the main navigation bar [Y/n]', True)
        if add_login_link:
            create_page(
                'Login', 'cms/home.html', initial_language,
                menu_title='Login', slug='login', overwrite_url=reverse('account_login'),
                in_navigation=True, limit_visibility_in_menu=VISIBILITY_ANONYMOUS, published=True
            )
            self.stdout.write('Login link added.\n')
            create_page(
                'My Account', 'cms/home.html', initial_language,
                menu_title='My Account', slug='profile', overwrite_url=reverse('accountProfile'),
                in_navigation=True, limit_visibility_in_menu=VISIBILITY_USERS, published=True
            )
            self.stdout.write('\'My Account\' link added.\n')
            create_page(
                'Logout', 'cms/home.html', initial_language,
                menu_title='Logout', slug='logout', overwrite_url=reverse('account_logout'),
                in_navigation=True, limit_visibility_in_menu=VISIBILITY_USERS, published=True
            )
            self.stdout.write('Logout link added.\n')

        if apps.is_installed('danceschool.payments.paypal'):
            call_command('setup_paypal')

        if apps.is_installed('danceschool.payments.stripe'):
            call_command('setup_stripe')

        if apps.is_installed('danceschool.payments.square'):
            call_command('setup_square')

        call_command('setup_permissions')

        # Finished with setup process
        self.stdout.write(self.style.SUCCESS('Successfully setup dance school!  Now enter \'python manage.py runserver\' command to test installation.'))



from django.conf.urls import url

from .views import InstructorPaymentsView, OtherInstructorPaymentsView, FinancesByMonthView, FinancesByEventView, AllExpensesViewCSV, AllRevenuesViewCSV, FinancialDetailView, ExpenseReportingView, RevenueReportingView, CompensationRuleUpdateView
from .ajax import updateEventRegistrations
from .autocomplete_light_registry import PaymentMethodAutoComplete

urlpatterns = [
    url(r'^instructor-payments/$', InstructorPaymentsView.as_view(), name='instructorPayments'),
    url(r'^instructor-payments/(?P<year>[\w\+]+)/$', InstructorPaymentsView.as_view(), name='instructorPayments'),
    url(r'^instructor-payments/(?P<year>[\w\+]+)/(?P<first_name>[\w\+\.]+)-(?P<last_name>[\w\+\.]+)/$', OtherInstructorPaymentsView.as_view(), name='instructorPayments'),
    url(r'^instructor-payments/csv/$', InstructorPaymentsView.as_view(as_csv=True), name='instructorPaymentsCSV'),
    url(r'^instructor-payments/(?P<year>[\w\+]+)/csv/$', InstructorPaymentsView.as_view(as_csv=True), name='instructorPaymentsCSV'),
    url(r'^instructor-payments/(?P<year>[\w\+]+)/(?P<first_name>[\w\+\.]+)-(?P<last_name>[\w\+\.]+)/csv/$', OtherInstructorPaymentsView.as_view(as_csv=True), name='instructorPaymentsCSV'),

    url(r'^submit-expenses/$', ExpenseReportingView.as_view(), name='submitExpenses'),
    url(r'^submit-revenues/$', RevenueReportingView.as_view(), name='submitRevenues'),

    # These URLs are for Ajax/autocomplete functionality
    url(r'^submit-revenues/eventfilter/$', updateEventRegistrations, name='ajaxhandler_updateEventRegistrations'),
    url(r'^autocomplete/paymentmethod/$', PaymentMethodAutoComplete.as_view(), name='paymentMethod-list-autocomplete'),

    # These URLs are for the financial views
    url(r'^finances/detail/(?P<year>[\w\+]+)/(?P<month>[\w\+]+)/$', FinancialDetailView.as_view(), name='financialDetailView'),
    url(r'^finances/detail/(?P<year>[\w\+]+)/$', FinancialDetailView.as_view(), name='financialDetailView'),
    url(r'^finances/detail/$', FinancialDetailView.as_view(), name='financialDetailView'),

    url(r'^finances/csv/$', FinancesByMonthView.as_view(as_csv=True), name='financesByMonthCSV'),
    url(r'^finances/(?P<year>[\w\+]+)/$', FinancesByMonthView.as_view(), name='financesByMonth'),
    url(r'^finances/(?P<year>[\w\+]+)/csv/$', FinancesByMonthView.as_view(as_csv=True), name='financesByMonthCSV'),
    url(r'^finances/$', FinancesByMonthView.as_view(), name='financesByMonth'),

    url(r'^finances-byevent/csv/$', FinancesByEventView.as_view(as_csv=True), name='financesByEventCSV'),
    url(r'^finances-byevent/(?P<year>[\w\+]+)/csv/$', FinancesByEventView.as_view(as_csv=True), name='financesByEventCSV'),
    url(r'^finances-byevent/$', FinancesByEventView.as_view(), name='financesByEvent'),

    url(r'^finances/expenses/(?P<year>[\w\+]+)/csv/$', AllExpensesViewCSV.as_view(), name='allexpensesCSV'),
    url(r'^finances/revenues/(?P<year>[\w\+]+)/csv/$', AllRevenuesViewCSV.as_view(), name='allrevenuesCSV'),

    url(r'^compensation/update/$', CompensationRuleUpdateView.as_view(), name='updateCompensationRules'),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-04-25 00:10
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import filer.fields.file


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('core', '0001_initial'),
        ('filer', '0007_auto_20161016_1055'),
        ('vouchers', '0001_initial'),
        ('financial', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name='revenueitem',
            name='purchasedVoucher',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='vouchers.Voucher', verbose_name='Purchased voucher/gift certificate'),
        ),
        migrations.AddField(
            model_name='revenueitem',
            name='registration',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='core.Registration'),
        ),
        migrations.AddField(
            model_name='revenueitem',
            name='submissionUser',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='revenuessubmittedby', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='expenseitem',
            name='attachment',
            field=filer.fields.file.FilerFileField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='expense_attachment', to='filer.File', verbose_name='Attach File (optional)'),
        ),
        migrations.AddField(
            model_name='expenseitem',
            name='category',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='financial.ExpenseCategory'),
        ),
        migrations.AddField(
            model_name='expenseitem',
            name='event',
            field=models.ForeignKey(blank=True, help_text='If this item is associated with an Event, enter it here.', null=True, on_delete=django.db.models.deletion.CASCADE, to='core.Event'),
        ),
        migrations.AddField(
            model_name='expenseitem',
            name='eventstaffmember',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='core.EventStaffMember'),
        ),
        migrations.AddField(
            model_name='expenseitem',
            name='eventvenue',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='venueexpense', to='core.Event'),
        ),
        migrations.AddField(
            model_name='expenseitem',
            name='payToLocation',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='core.Location'),
        ),
        migrations.AddField(
            model_name='expenseitem',
            name='payToUser',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='payToUser', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='expenseitem',
            name='submissionUser',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='expensessubmittedby', to=settings.AUTH_USER_MODEL),
        ),
    ]



from django.db import models
from django.core.validators import MinValueValidator
from django.utils.translation import ugettext_lazy as _

from djchoices import DjangoChoices, ChoiceItem
from calendar import day_name
from collections import namedtuple

from danceschool.core.models import PricingTier, DanceTypeLevel, Registration, TemporaryRegistration, Customer, CustomerGroup


class PointGroup(models.Model):
    name = models.CharField(_('Name'),max_length=50,unique=True,help_text=_('Give this pricing tier point group a name (e.g. \'Regular Series Class Hours\')'))

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = _('Point group type')
        verbose_name_plural = _('Point group types')


class PricingTierGroup(models.Model):
    '''
    Because we want to keep the discounts app completely separable from the core app
    for schools that do not use discounting, this model serves as a go-between
    from the PricingTier objects of the core app to the PointGroup objects used
    in constructing discounts.
    '''
    group = models.ForeignKey(PointGroup,verbose_name=_('Point group'))
    pricingTier = models.OneToOneField(PricingTier,verbose_name=_('Pricing tier'))
    points = models.PositiveIntegerField(_('# of Points'),default=0)

    def __str__(self):
        return str(_('%s points for pricing tier %s' % (self.group.name, self.pricingTier.name)))

    class Meta:
        ''' Only one pricingtiergroup per pricingtier and group combo '''
        unique_together = (('pricingTier','group'),)
        verbose_name = _('Pricing tier point group')
        verbose_name_plural = _('Pricing tier point groups')


class DiscountCategory(models.Model):
    '''
    The discounts app allows one discount to be applied per category
    '''
    name = models.CharField(_('Category name'),max_length=30,unique=True)
    order = models.FloatField(
        _('Category order'),
        help_text=_('Discounts from categories with lower numbers are applied first.'),
        unique=True
    )
    cannotCombine = models.BooleanField(
        _('Cannot be combined'),
        help_text=_('If checked, then discounts in this category cannot be combined with any other type of discount.'),
        default=False,
    )

    def __str__(self):
        return self.name

    class Meta:
        ordering = ('order',)
        verbose_name = _('Discount category')
        verbose_name_plural = _('Discount categories')


class DiscountCombo(models.Model):

    # Choices of Discount Types
    class DiscountType(DjangoChoices):
        flatPrice = ChoiceItem('F',_('Exact Specified Price'))
        dollarDiscount = ChoiceItem('D',_('Dollar Discount from Regular Price'))
        percentDiscount = ChoiceItem('P',_('Percentage Discount from Regular Price'))
        addOn = ChoiceItem('A',_('Free Add-on Item (Can be combined with other discounts)'))

    # An applied discount may not apply to all items in a customer's
    # cart, so an instance of this class keeps track of the code
    # as well as the items applicable.  ItemTuples should be a series of 2-tuples.
    # For point-based discounts, the second position in the tuple represents what fraction of the
    # points were used from this item to match the discount (1 if the item was counted in its
    # entirety toward the discount). In cases where a discount does not count against an item entirely
    # (e.g. the 9 hours of class discount applied to three four-week classes), the full price for that
    # item will be calculated by multiplying the remaining fraction of points for that item (e.g. 75%
    # in the above example) against the regular (non-student, online registration) price for that item.
    # In practice, this means that a discount on the whole item will never be superceded by a discount
    # on a portion of the item as long as the percentage discounts for more point are always larger
    # than the discounts for fewer points (e.g. additional hours of class are cheaper).
    ApplicableDiscountCode = namedtuple('ApplicableDiscountCode',['code','items','itemTuples'])

    # Net allocated prices are optional, everything else is required.
    DiscountInfo = namedtuple('DiscountInfo',['code','net_price','discount_amount','net_allocated_prices'])
    DiscountInfo.__new__.__defaults__ = ([],)

    # A DiscountApplication contains a list of DiscountInfo tuples to be applied, along with an optional
    # total for ineligible items to be added to the discounted net price at the end.
    DiscountApplication = namedtuple('DiscountApplication',['items','ineligible_total'])
    DiscountApplication.__new__.__defaults__ = ([],0)

    name = models.CharField(_('Name'),max_length=50,unique=True,help_text=_('Give this discount combination a name (e.g. \'Two 4-week series\')'))
    active = models.BooleanField(_('Active'),default=True,help_text=_('Uncheck this to \'turn off\' this discount'))
    category = models.ForeignKey(
        DiscountCategory,verbose_name=_('Discount category'),
        help_text=_('One discount can be applied per category, and the order in which discounts are applied depends on the order parameter of the categories.')
    )

    # If null, then there is no expiration date
    expirationDate = models.DateTimeField(_('Expiration Date'), null=True,blank=True,help_text=_('Leave blank for no expiration.'))

    studentsOnly = models.BooleanField(verbose_name=_('Discount for HS/college/university students only'),default=False,help_text=_('Check this box to create student discounts.'))
    newCustomersOnly = models.BooleanField(verbose_name=_('Discount for New Customers only'),default=False)
    daysInAdvanceRequired = models.PositiveIntegerField(
        _('Must register __ days in advance'),
        null=True, blank=True,
        help_text=_(
            'For this discount to apply, all components must be satisfied by events that begin '
            'this many days in the future (measured from midnight of the date of registration). '
            'Leave blank for no restriction.'
        ),
    )
    firstXRegistered = models.PositiveSmallIntegerField(
        _('Only for first __ registrants'),
        null=True, blank=True,
        help_text=_(
            'For this discount to apply, all components must be satisfied by events that have fewer '
            'than this many individuals registered (including registrations in progress). '
            'Only one discount per category can be applied, so if you define tiered discounts '
            'in the same category, only the best available discount will be used.'
        )
    )

    discountType = models.CharField(_('Discount type'),max_length=1,help_text=_('Is this a flat price, a dollar amount discount, a \'percentage off\' discount, or a free add-on?'),choices=DiscountType.choices,default=DiscountType.dollarDiscount)

    # For flat price discounts
    onlinePrice = models.FloatField(_('Online price'),null=True,blank=True,validators=[MinValueValidator(0)])
    doorPrice = models.FloatField(_('At-the-door price'),null=True,blank=True,validators=[MinValueValidator(0)])

    # For 'dollars off' discounts
    dollarDiscount = models.FloatField(_('Amount of dollar discount'),null=True,blank=True,help_text=_('This amount will be subtracted from the customer\'s total (in currency units, e.g. dollars).'),validators=[MinValueValidator(0)])

    # For 'percentage off' discounts
    percentDiscount = models.FloatField(_('Amount of percentage discount'),null=True,blank=True,help_text=_('This percentage will be subtracted from the customer\'s total.'),validators=[MinValueValidator(0)])
    percentUniversallyApplied = models.BooleanField(_('Percentage discount is univerally applied'),default=False,help_text=_('If checked, then the percentage discount will be applied to all items in the order, not just the items that qualify the order for this discount combination (e.g. 20% off all registrations of three or more classes.'))

    def applyAndAllocate(self,allocatedPrices,tieredTuples,payAtDoor=False):
        '''
        This method takes an initial allocation of prices across events, and
        an identical length list of allocation tuples.  It applies the rule
        specified by this discount, allocates the discount across the listed
        items, and returns both the price and the allocation
        '''
        initial_net_price = sum([x for x in allocatedPrices])

        if self.discountType == self.DiscountType.flatPrice:
            # Flat-price for all applicable items (partial application for items which are
            # only partially needed to apply the discount).  Flat prices ignore any previous discounts
            # in other categories which may have been the best, but they only are applied if they are
            # lower than the price that would be feasible by applying those prior discounts alone.
            applicable_price = self.getFlatPrice(payAtDoor) or 0

            this_price = applicable_price \
                + sum([x[0].event.getBasePrice(payAtDoor=payAtDoor) * x[1] if x[1] != 1 else x[0].price for x in tieredTuples])

            # Flat prices are allocated equally across all events
            this_allocated_prices = [x * (this_price / initial_net_price) for x in allocatedPrices]

        elif self.discountType == self.DiscountType.dollarDiscount:
            # Discount the set of applicable items by a specific number of dollars (currency units)
            # Dollar discounts are allocated equally across all events.
            this_price = initial_net_price - self.dollarDiscount
            this_allocated_prices = [x * (this_price / initial_net_price) for x in allocatedPrices]

        elif self.discountType == DiscountCombo.DiscountType.percentDiscount:
            # Percentage off discounts, which may be applied to all items in the cart,
            # or just to the items that were needed to apply the discount

            if self.percentUniversallyApplied:
                this_price = \
                    initial_net_price * (1 - (max(min(self.percentDiscount or 0,100),0) / 100))
                this_allocated_prices = [x * (this_price / initial_net_price) for x in allocatedPrices]
            else:
                # Allocate the percentage discount based on the prior allocation from the prior category
                this_price = 0
                this_allocated_prices = []

                for idx, val in enumerate(tieredTuples):
                    this_val = (
                        allocatedPrices[idx] *
                        (1 - val[1]) * (1 - (max(min(self.percentDiscount or 0,100),0) / 100)) +
                        allocatedPrices[idx] * val[1]
                    )
                    this_allocated_prices.append(this_val)
                    this_price += this_val
        else:
            raise KeyError(_('Invalid discount type.'))

        if this_price < initial_net_price:
            # Ensure no negative prices
            this_price = max(this_price, 0)
            return self.DiscountInfo(self, this_price, initial_net_price - this_price, this_allocated_prices)

    def getFlatPrice(self,payAtDoor=False):
        '''
        Rather than embedding logic re: door pricing,
        other code can call this method.
        '''
        if self.discountType is not DiscountCombo.DiscountType.flatPrice:
            return None
        if payAtDoor:
            return self.doorPrice
        else:
            return self.onlinePrice

    def getComponentList(self):
        '''
        This function just returns a list with items that are supposed to
        be present in the the list multiple times as multiple elements
        of the list.  It simplifies checking whether a discount's conditions
        are satisfied.
        '''

        component_list = []

        for x in self.discountcombocomponent_set.all():
            for y in range(0,x.quantity):
                component_list += [x]

        component_list.sort(key=lambda x: x.quantity, reverse=True)
        return component_list

    def save(self, *args, **kwargs):
        '''
        Don't save any passed values related to a type of discount
        that is not the specified type
        '''

        if self.discountType != self.DiscountType.flatPrice:
            self.onlinePrice = None
            self.doorPrice = None

        if self.discountType != self.DiscountType.dollarDiscount:
            self.dollarDiscount = None

        if self.discountType != self.DiscountType.percentDiscount:
            self.percentDiscount = None
            self.percentUniversallyApplied = False

        super(DiscountCombo, self).save(*args, **kwargs)

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = _('Discount')
        verbose_name_plural = _('Discounts')


class DiscountComboComponent(models.Model):

    # For weekday-specific discounts
    DAYS_CHOICES = [(None,''),] + [(k,x) for k,x in enumerate(day_name)]

    discountCombo = models.ForeignKey(DiscountCombo,verbose_name=_('Discount'))

    pointGroup = models.ForeignKey(PointGroup,verbose_name=_('Pricing Tier Point Group'))

    allWithinPointGroup = models.BooleanField(verbose_name=_('Applies to all within Point Group'),help_text=_('If checked, then this discount applies to this quantity or more within the point group.  Use, for example, for all-in passes.'))

    quantity = models.PositiveSmallIntegerField(_('Quantity'),default=1)
    level = models.ForeignKey(DanceTypeLevel,null=True,blank=True,verbose_name=_('Dance type/level'))
    weekday = models.PositiveSmallIntegerField(_('Weekday'),choices=DAYS_CHOICES,blank=True,null=True,help_text=_('Leave this blank for no restriction on days of the week'))

    class Meta:
        verbose_name = _('Required component of discount')
        verbose_name_plural = _('Required components of discount')


class TemporaryRegistrationDiscount(models.Model):
    registration = models.ForeignKey(TemporaryRegistration,verbose_name=_('Temporary registration'))
    discount = models.ForeignKey(DiscountCombo,verbose_name=_('Discount'))
    discountAmount = models.FloatField(verbose_name=_('Amount of discount'),validators=[MinValueValidator(0)])

    class Meta:
        unique_together = ('registration', 'discount')
        verbose_name = _('Discount applied to temporary registration')
        verbose_name_plural = _('Discounts applied to temporary registrations')


class RegistrationDiscount(models.Model):
    registration = models.ForeignKey(Registration,verbose_name=_('Registration'))
    discount = models.ForeignKey(DiscountCombo,verbose_name=_('Discount'))
    discountAmount = models.FloatField(verbose_name=_('Amount of discount'),validators=[MinValueValidator(0)])

    class Meta:
        unique_together = ('registration', 'discount')
        verbose_name = _('Discount applied to registration')
        verbose_name_plural = _('Discounts applied to registrations')


class CustomerGroupDiscount(models.Model):
    ''' Some discounts are only available for specific customer groups '''
    group = models.ForeignKey(CustomerGroup,verbose_name=_('Customer group'))
    discountCombo = models.ForeignKey(DiscountCombo,verbose_name=_('Discount'))

    class Meta:
        verbose_name = _('Group-specific discount restriction')
        verbose_name_plural = _('Group-specific discount restrictions')


class CustomerDiscount(models.Model):
    ''' Some discounts are only available for specific customers. '''
    customer = models.ForeignKey(Customer,verbose_name=_('Customer'))
    discountCombo = models.ForeignKey(DiscountCombo,verbose_name=_('Discount'))

    class Meta:
        verbose_name = _('Customer-specific discount restriction')
        verbose_name_plural = _('Customer-specific discount restrictions')



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-07-17 16:42
from __future__ import unicode_literals

import datetime
from django.db import migrations, models
import djangocms_text_ckeditor.fields


class Migration(migrations.Migration):

    dependencies = [
        ('news', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='newsitem',
            options={'ordering': ('-pinThis', '-publicationDate'), 'verbose_name': 'News item', 'verbose_name_plural': 'News items'},
        ),
        migrations.AlterField(
            model_name='latestnewspluginmodel',
            name='template',
            field=models.CharField(blank=True, max_length=250, null=True, verbose_name='Template'),
        ),
        migrations.AlterField(
            model_name='newsitem',
            name='alert',
            field=models.BooleanField(default=False, help_text='Alerts (such as cancellations) may be displayed in an emphasized fashion.', verbose_name='Alert'),
        ),
        migrations.AlterField(
            model_name='newsitem',
            name='content',
            field=djangocms_text_ckeditor.fields.HTMLField(help_text='Insert news content here.', verbose_name='Content'),
        ),
        migrations.AlterField(
            model_name='newsitem',
            name='creationDate',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Creation date'),
        ),
        migrations.AlterField(
            model_name='newsitem',
            name='draft',
            field=models.BooleanField(default=False, help_text='Check to hide from publication', verbose_name='Draft status'),
        ),
        migrations.AlterField(
            model_name='newsitem',
            name='modifiedDate',
            field=models.DateTimeField(auto_now=True, help_text='Time of most recent edit', verbose_name='Last modified date'),
        ),
        migrations.AlterField(
            model_name='newsitem',
            name='publicationDate',
            field=models.DateTimeField(default=datetime.datetime.now, verbose_name='Publication date'),
        ),
        migrations.AlterField(
            model_name='newsitem',
            name='title',
            field=models.CharField(help_text='Title the news item. Make sure this is descriptive.', max_length=200, verbose_name='Title'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-08-21 16:26
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0013_auto_20170815_2235'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='emailtemplate',
            options={'ordering': ('name',), 'permissions': (('send_email', 'Can send emails using the SendEmailView'),), 'verbose_name': 'Email template', 'verbose_name_plural': 'Email templates'},
        ),
        migrations.AlterModelOptions(
            name='eventoccurrence',
            options={'ordering': ('event', 'startTime'), 'verbose_name': 'Event occurrence', 'verbose_name_plural': 'Event occurrences'},
        ),
        migrations.AlterModelOptions(
            name='eventstaffcategory',
            options={'ordering': ('name',), 'verbose_name': 'Event staff category', 'verbose_name_plural': 'Event staff categories'},
        ),
        migrations.AlterModelOptions(
            name='eventstaffmember',
            options={'ordering': ('event', 'staffMember__lastName', 'staffMember__firstName'), 'verbose_name': 'Event staff member', 'verbose_name_plural': 'Event staff members'},
        ),
        migrations.AlterModelOptions(
            name='invoice',
            options={'ordering': ('-modifiedDate',), 'permissions': (('view_all_invoices', 'Can view invoices without passing the validation string.'), ('send_invoices', 'Can send invoices to students requesting payment'), ('process_refunds', 'Can refund customers for registrations and other invoice payments.')), 'verbose_name': 'Invoice', 'verbose_name_plural': 'Invoices'},
        ),
        migrations.AlterModelOptions(
            name='paymentrecord',
            options={'ordering': ('-modifiedDate',), 'verbose_name': 'Payment record', 'verbose_name_plural': 'Payment records'},
        ),
        migrations.AlterModelOptions(
            name='pricingtier',
            options={'ordering': ('name',), 'verbose_name': 'Pricing tier', 'verbose_name_plural': 'Pricing tiers'},
        ),
        migrations.AlterModelOptions(
            name='staffmember',
            options={'ordering': ('lastName', 'firstName'), 'permissions': (('view_staff_directory', 'Can access the staff directory view'), ('view_school_stats', "Can view statistics about the school's performance.")), 'verbose_name': 'Staff member', 'verbose_name_plural': 'Staff members'},
        ),
    ]



from django.utils.translation import ugettext_lazy as _

from danceschool.core.registries import plugin_templates_registry, PluginTemplateBase


@plugin_templates_registry.register
class VoucherStatsTemplate(PluginTemplateBase):
    template_name = 'stats/schoolstats_voucherusage.html'
    plugin = 'StatsGraphPlugin'
    description = _('Statistics on Usage of Vouchers')



'''
This file defines a variety of preferences that must be set in the DB,
but can be changed dynamically.
'''

from django.utils.translation import ugettext_lazy as _

from dynamic_preferences.types import FloatPreference, Section
from dynamic_preferences.registries import global_preferences_registry

paypal = Section('paypal',_('Paypal'))


##############################
# Referral Program Preferences
#
@global_preferences_registry.register
class FixedTransactionFee(FloatPreference):
    section = paypal
    name = 'fixedTransactionFee'
    verbose_name = _('Paypal fixed transaction fee')
    help_text = _('Paypal fees consist of a fixed fee plus a percentage of the total price.  When a refund is made, the  the U.S., the variable percentage is refunded, but the fixed fee is not.  In the U.S., this fixed fee is $0.30.  Entering your country\'s fee here ensures that your accounting will remain accurate when you refund students.')
    default = 0.30



from django.utils.translation import ugettext_lazy as _

from dynamic_preferences.types import BooleanPreference, StringPreference, Section
from dynamic_preferences.registries import global_preferences_registry

# we create some section objects to link related preferences together

registration = Section('registration', _('Registration'))


@global_preferences_registry.register
class BanListEnabled(BooleanPreference):
    section = registration
    name = 'enableBanList'
    verbose_name = _('Enable Automated Enforcement of Ban List')
    help_text = _(
        'If this box is checked, then if an individual whose name or email ' +
        'address appears on the ban list will be prevented from registration ' +
        'and asked to contact the school.  Their attempt to register will be logged.'
    )
    default = True


@global_preferences_registry.register
class BanListContactEmail(StringPreference):
    section = registration
    name = 'banListContactEmail'
    verbose_name = _(
        'Individuals who are prevented from registration are asked ' +
        'to contact this address.'
    )
    help_text = _('If no email is specified, then the default school email will be used.')
    default = ''


@global_preferences_registry.register
class BanListNotificationEmail(StringPreference):
    section = registration
    name = 'banListNotificationEmail'
    verbose_name = _(
        'Send a notification of attempted registration by banned users to ' +
        'this address (optional).'
    )
    help_text = _('If no email is specified, then no notification will be sent.')
    default = ''



from django import forms
from django.utils.translation import ugettext_lazy as _


class VoucherCustomizationForm(forms.Form):

    emailTo = forms.EmailField(label=_('Email gift certificate to'),required=True)
    emailType = forms.ChoiceField(
        label=_('This email belongs to'),
        widget=forms.RadioSelect,
        required=True,
        choices=[('R',_('The recipient')),('P',_('The purchaser'))],
        help_text=_('If emailing directly to the recipient, a gift message will be sent. If emailing to the purchaser, the gift message will only be included as a PDF attachment.')
    )

    recipientName = forms.CharField(label=_('Recipient name (optional)'),required=False,help_text=_('If provided, this will be used to customize the gift message.'))
    fromName = forms.CharField(label=_('From name (optional)'),required=False)

    message = forms.CharField(label=_('Enter a message to the recipient (optional)'),required=False)



from django.core.urlresolvers import reverse
from django.utils.translation import ugettext_lazy as _

from cms.toolbar_pool import toolbar_pool
from cms.toolbar_base import CMSToolbar
from cms.toolbar.items import Break


@toolbar_pool.register
class PaypalLinksToolbar(CMSToolbar):
    ''' Add Paypal items to the financial menu '''

    def populate(self):
        if not (self.request.user.has_perm('paypal.change_paypalpaymentrecord')):
            return

        financial_menu = self.toolbar.get_or_create_menu(
            'financial', _('Finances'))

        position = financial_menu.find_first(
            Break,
            identifier='financial_related_items_break'
        )

        if not position:
            financial_menu.add_break('financial_related_items_break')
            position = financial_menu.find_first(
                Break,
                identifier='financial_related_items_break'
            ) + 1

        related_menu = financial_menu.get_or_create_menu('financial-related',_('Related Items'), position=position)

        if self.request.user.has_perm('paypal.change_paypalpaymentrecord'):
            related_menu.add_link_item(_('Paypal Payment Records'), url=reverse('admin:paypal_paypalpaymentrecord_changelist'))



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-07-17 16:42
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('prerequisites', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='customerrequirement',
            options={'verbose_name': 'Customer-level requirement record', 'verbose_name_plural': 'Customer-level requirement records'},
        ),
        migrations.AlterModelOptions(
            name='requirement',
            options={'permissions': (('ignore_requirements', 'Can register users for series regardless of any prerequisites or requirements'),), 'verbose_name': 'Class requirements'},
        ),
        migrations.AlterModelOptions(
            name='requirementitem',
            options={'verbose_name': 'Requirement item', 'verbose_name_plural': 'Requirement items'},
        ),
        migrations.AlterField(
            model_name='customerrequirement',
            name='customer',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Customer', verbose_name='Customer'),
        ),
        migrations.AlterField(
            model_name='customerrequirement',
            name='modifiedDate',
            field=models.DateTimeField(auto_now=True, verbose_name='Last modified date'),
        ),
        migrations.AlterField(
            model_name='customerrequirement',
            name='requirement',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='prerequisites.Requirement', verbose_name='Requirement'),
        ),
        migrations.AlterField(
            model_name='customerrequirement',
            name='role',
            field=models.ForeignKey(blank=True, help_text='Role must be specified only for requirements for which roles are enforced.', null=True, on_delete=django.db.models.deletion.CASCADE, to='core.DanceRole', verbose_name='Dance role'),
        ),
        migrations.AlterField(
            model_name='customerrequirement',
            name='submissionDate',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Submission date'),
        ),
        migrations.AlterField(
            model_name='requirementitem',
            name='requiredClass',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='core.ClassDescription', verbose_name='Required class'),
        ),
        migrations.AlterField(
            model_name='requirementitem',
            name='requiredLevel',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='core.DanceTypeLevel', verbose_name='Required Dance type/level'),
        ),
        migrations.AlterField(
            model_name='requirementitem',
            name='requirement',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='prerequisites.Requirement', verbose_name='Requirement'),
        ),
    ]



from django.contrib import admin
from django.forms import ModelForm, HiddenInput
from django.utils.translation import ugettext_lazy as _

from .models import InstructorPrivateLessonDetails, InstructorAvailabilitySlot, PrivateLessonEvent, PrivateLessonCustomer

from danceschool.core.models import Instructor, EventStaffMember
from danceschool.core.admin import EventChildAdmin, EventOccurrenceInline, EventRegistrationInline
from danceschool.core.constants import getConstant
from danceschool.core.forms import LocationWithDataWidget


class InstructorPrivateLessonDetailsInline(admin.TabularInline):
    model = InstructorPrivateLessonDetails
    extra = 0

    # Prevents adding new voucher uses without going through
    # the standard registration process
    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(InstructorAvailabilitySlot)
class InstructorAvailabilitySlotAdmin(admin.ModelAdmin):
    exclude = []


class PrivateLessonCustomerInline(admin.StackedInline):
    model = PrivateLessonCustomer
    extra = 0


class PrivateLessonTeacherInlineForm(ModelForm):
    def __init__(self, *args, **kwargs):
        super(PrivateLessonTeacherInlineForm,self).__init__(*args,**kwargs)

        self.fields['staffMember'].label = _('Instructor')
        self.fields['category'].initial = getConstant('privateLessons__eventStaffCategoryPrivateLesson').id

        # Impose restrictions on new records, but not on existing ones.
        if not kwargs.get('instance',None):
            # Filter out retired teachers
            self.fields['staffMember'].queryset = Instructor.objects.exclude(status__in=[Instructor.InstructorStatus.retired,Instructor.InstructorStatus.hidden,Instructor.InstructorStatus.retiredGuest])
        else:
            self.fields['staffMember'].queryset = Instructor.objects.all()

        self.fields['staffMember'].queryset = self.fields['staffMember'].queryset.order_by('status','firstName','lastName')

    class Meta:
        widgets = {
            'category': HiddenInput(),
        }


class PrivateLessonTeacherInline(admin.StackedInline):
    model = EventStaffMember
    form = PrivateLessonTeacherInlineForm
    exclude = ('replacedStaffMember','occurrences','submissionUser')
    extra = 0

    def save_model(self,request,obj,form,change):
        obj.replacedStaffMember = None
        obj.occurrences = obj.event.eventoccurrence_set.all()
        obj.submissionUser = request.user
        obj.save()


class PrivateLessonEventRegistrationInline(EventRegistrationInline):
    ''' View/edit but do not add/delete EventRegistrations from here. '''
    fields = ['role','cancelled','price','netPrice']
    readonly_fields = ['price','netPrice']


class PrivateLessonEventAdminForm(ModelForm):
    '''
    Custom form for private lesson events is needed to include necessary
    Javascript for room selection, even though capacity is not
    an included field in this admin.
    '''

    class Meta:
        model = PrivateLessonEvent
        exclude = ['month','year','startTime','endTime','duration','submissionUser','registrationOpen','capacity','status']
        widgets = {
            'location': LocationWithDataWidget,
        }

    class Media:
        js = ('js/serieslocation_capacity_change.js','js/location_related_objects_lookup.js')


@admin.register(PrivateLessonEvent)
class PrivateLessonEventAdmin(EventChildAdmin):
    base_model = PrivateLessonEvent
    form = PrivateLessonEventAdminForm
    show_in_index = True

    list_display = ('teacherNames','customerNames','startTime','durationMinutes','location','pricingTier')
    list_filter = ('location','room','startTime','pricingTier')

    fieldsets = (
        (None, {'fields': (('location','room'),'pricingTier','participants','comments',)}),
    )

    def teacherNames(self,obj):
        return ', '.join([x.staffMember.fullName for x in obj.eventstaffmember_set.all()])
    teacherNames.short_description = _('Instructors')

    def customerNames(self,obj):
        return ', '.join([x.fullName for x in obj.customers])
    customerNames.short_description = _('Customers')

    inlines = [EventOccurrenceInline,PrivateLessonCustomerInline,PrivateLessonTeacherInline,PrivateLessonEventRegistrationInline]


admin.site._registry[Instructor].inlines.insert(0,InstructorPrivateLessonDetailsInline)



from django.conf.urls import include, url
from django.contrib.auth.decorators import user_passes_test
from django.contrib.sitemaps.views import sitemap
from django.apps import apps

from cms.sitemaps import CMSSitemap

from dynamic_preferences.views import PreferenceFormView
from dynamic_preferences.registries import global_preferences_registry
from dynamic_preferences.forms import GlobalPreferenceForm

from danceschool.core.sitemaps import EventSitemap

urlpatterns = [
    # For site configuration
    url(r'^settings/global/$',
        user_passes_test(lambda u: u.is_superuser)(PreferenceFormView.as_view(
            registry=global_preferences_registry,
            form_class=GlobalPreferenceForm,
            template_name='dynamic_preferences/form_danceschool.html',)),
        name="dynamic_preferences.global"),

    url(r'^settings/global/(?P<section>[\w\ ]+)$',
        user_passes_test(lambda u: u.is_superuser)(PreferenceFormView.as_view(
            registry=global_preferences_registry,
            form_class=GlobalPreferenceForm,
            template_name='dynamic_preferences/form_danceschool.html',)),
        name="dynamic_preferences.global.section"),

    # For Django-filer
    url(r'^filer/', include('filer.urls')),
    url(r'^', include('filer.server.urls')),

    # For Django-filer in CKeditor
    url(r'^filebrowser_filer/', include('ckeditor_filebrowser_filer.urls')),

    # For automatically-generated sitemaps
    url(r'^sitemap\.xml$',
        sitemap,
        {'sitemaps': {'event': EventSitemap, 'page': CMSSitemap}},
        name='django.contrib.sitemaps.views.sitemap'),

    # For better authentication
    url(r'^accounts/', include('allauth.urls')),

    # For Django CMS forms
    url(r'^', include('djangocms_forms.urls')),
    
    # The URLS associated with all built-in core functionality.
    url(r'^', include('danceschool.core.urls')),
    url(r'^register/', include('danceschool.core.urls_registration')),
]

# If additional danceschool apps are installed, automatically add those URLs as well.
if apps.is_installed('danceschool.banlist'):
    urlpatterns.append(url(r'^banlist/', include('danceschool.banlist.urls')),)

if apps.is_installed('danceschool.discounts'):
    urlpatterns.append(url(r'^discounts/', include('danceschool.discounts.urls')),)

if apps.is_installed('danceschool.financial'):
    urlpatterns.append(url(r'^financial/', include('danceschool.financial.urls')),)

if apps.is_installed('danceschool.prerequisites'):
    urlpatterns.append(url(r'^prerequisites/', include('danceschool.prerequisites.urls')),)

if apps.is_installed('danceschool.private_events'):
    urlpatterns.append(url(r'^private-events/', include('danceschool.private_events.urls')),)

if apps.is_installed('danceschool.private_lessons'):
    urlpatterns.append(url(r'^private-lessons/', include('danceschool.private_lessons.urls')),)

if apps.is_installed('danceschool.vouchers'):
    urlpatterns.append(url(r'^vouchers/', include('danceschool.vouchers.urls')),)

if apps.is_installed('danceschool.payments.paypal'):
    urlpatterns.append(url(r'^paypal/', include('danceschool.payments.paypal.urls')),)

if apps.is_installed('danceschool.payments.stripe'):
    urlpatterns.append(url(r'^stripe/', include('danceschool.payments.stripe.urls')),)

if apps.is_installed('danceschool.payments.square'):
    urlpatterns.append(url(r'^square/', include('danceschool.payments.square.urls')),)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-07-24 21:14
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('discounts', '0002_auto_20170717_1642'),
    ]

    operations = [
        migrations.AlterField(
            model_name='discountcombo',
            name='discountType',
            field=models.CharField(choices=[('F', 'Exact Specified Price'), ('D', 'Dollar Discount from Regular Price'), ('P', 'Percentage Discount from Regular Price'), ('A', 'Free Add-on Item (Can be combined with other discounts)')], default='D', help_text="Is this a flat price, a dollar amount discount, a 'percentage off' discount, or a free add-on?", max_length=1, verbose_name='Discount type'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-05-12 22:07
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('cms', '0016_auto_20160608_1535'),
    ]

    operations = [
        migrations.CreateModel(
            name='StatsGraphPluginModel',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='stats_statsgraphpluginmodel', serialize=False, to='cms.CMSPlugin')),
                ('template', models.CharField(blank=True, max_length=250, null=True, verbose_name='Template')),
            ],
            options={
                'abstract': False,
            },
            bases=('cms.cmsplugin',),
        ),
    ]



'''
This file contains helper functions used at the time that various apps are loaded
'''

import sys


def isPreliminaryRun():
    '''
    Check the arguments passed at runtime to ensure that this
    program is not being run to perform migrations or load data.
    '''
    prelim_params = ['loaddata','makemigrations','migrate']
    for param in prelim_params:
        if param in sys.argv:
            return True
    return False



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-07-21 10:53
from __future__ import unicode_literals

import cms.models.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('cms', '0016_auto_20160608_1535'),
        ('core', '0011_auto_20170720_1945'),
    ]

    operations = [
        migrations.CreateModel(
            name='StripeCharge',
            fields=[
                ('paymentrecord_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.PaymentRecord')),
                ('chargeId', models.CharField(max_length=32, verbose_name='Charge Id')),
                ('status', models.CharField(blank=True, max_length=30, null=True, verbose_name='Current status')),
            ],
            options={
                'verbose_name_plural': 'Stripe charges',
                'verbose_name': 'Stripe charge',
            },
            bases=('core.paymentrecord',),
        ),
        migrations.CreateModel(
            name='StripeChargeFormModel',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='stripe_stripechargeformmodel', serialize=False, to='cms.CMSPlugin')),
                ('defaultAmount', models.FloatField(default=0, help_text='The initial value for gift certificate forms.', verbose_name='Default amount')),
                ('successPage', cms.models.fields.PageField(help_text='When the user returns to the site after a successful transaction, send them to this page.', on_delete=django.db.models.deletion.CASCADE, related_name='successPageForStripe', to='cms.Page', verbose_name='Success Page')),
            ],
            options={
                'abstract': False,
            },
            bases=('cms.cmsplugin',),
        ),
    ]



from django.conf.urls import url
from .views import processSquarePayment, processPointOfSalePayment

urlpatterns = [
    url(r'^create_payment/$', processSquarePayment, name='processSquarePayment'),
    url(r'^process_pointofsale/$', processPointOfSalePayment, name='processSquarePointOfSale'),
]



from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.utils import timezone
from django.core.exceptions import ValidationError

from djchoices import DjangoChoices, ChoiceItem

from danceschool.core.models import Customer, ClassDescription, DanceTypeLevel, DanceRole, Registration, TemporaryRegistration, EventRegistration


class Requirement(models.Model):
    '''
    Requirements apply to either a ClassDescription (i.e. a specific class), or to a DanceTypeLevel (i.e. a specific level).
    They typically consist of one or more items, though if a Requirement has no items, then the only way it can be met is for
    a Customer to have explicitly met the requirement through a CustomerRequirement object.
    '''
    class BooleanChoice(DjangoChoices):
        booleanAnd = ChoiceItem('&',_('Must meet all requirement items'))
        booleanOr = ChoiceItem('|',_('Must meet one or more requirement items'))
        booleanNot = ChoiceItem('!',_('Must not meet any requirement items'))

    class EnforcementChoice(DjangoChoices):
        none = ChoiceItem('N',_('Enforcement disabled'))
        warning = ChoiceItem('W',_('Allow registration with warning'))
        error = ChoiceItem('E',_('Raise error and do not allow registration'))

    name = models.CharField(_('Requirement name/description'),max_length=300,help_text=_('If a customer does not meet the requirement for a series, then this description will be used to explain the issue (e.g. \'Must have taken Lindy 1 to take Lindy 2\').'))

    applicableClass = models.ForeignKey(ClassDescription,verbose_name=_('Applies to class'),null=True,blank=True)
    applicableLevel = models.ForeignKey(DanceTypeLevel,verbose_name=_('Applies to dance type/level'),null=True,blank=True)

    booleanRule = models.CharField(
        max_length=3,choices=BooleanChoice.choices,default=BooleanChoice.booleanAnd,
        verbose_name=_('How many items of this requirement must be met'),
        help_text=_('If you select an option other than \'Must meet all requirements\', then you can still enforce multiple requirements by adding another Requirement item.'))
    enforcementMethod = models.CharField(_('Enforcement method'),max_length=1,choices=EnforcementChoice.choices,default=EnforcementChoice.warning)

    applicableRole = models.ForeignKey(DanceRole,verbose_name=_('Applies only to role (optional)'),help_text=_('Most requirements apply identically to all dance roles.  If this requirement does not, then specify the role to which it applies here.'),null=True,blank=True)
    roleEnforced = models.BooleanField(_('Same dance role enforced'),default=False,help_text=_('If checked, then in order to meet the requirement, the customer must have met all individual requirements in the same dance role, and that must be the dance role for which they are registering.'))

    submissionDate = models.DateTimeField(_('Submission date'),auto_now_add=True)
    modifiedDate = models.DateTimeField(_('Last modified date'),auto_now=True)

    @property
    def enabled(self):
        return self.enforcementMethod != self.EnforcementChoice.none
    enabled.fget.short_description = _('Enabled')

    def customerMeetsRequirement(self, customer, danceRole=None, registration=None):
        '''
        This method checks whether a given customer meets a given set of requirements.
        '''

        cust_reqs = self.customerrequirement_set.filter(customer=customer,met=True)
        if customer:
            cust_priors = customer.eventregistration_set.filter(event__series__isnull=False)
        else:
            cust_priors = EventRegistration.objects.none()

        # If there's an explicit object stating that this customer meets the requirement, then we're done.
        if self.roleEnforced and danceRole and cust_reqs.filter(role=danceRole).exists():
            return True
        elif not self.roleEnforced and cust_reqs.exists():
            return True
        elif self.roleEnforced and not danceRole:
            return False

        # Go through each item for this requirement and see if the customer meets that item
        for item in self.requirementitem_set.all():
            filter_dict = {}
            if item.requiredLevel:
                filter_dict['event__series__classDescription__danceTypeLevel'] = item.requiredLevel
            if item.requiredClass:
                filter_dict['event__series__classDescription'] = item.requiredClass
            if self.roleEnforced:
                filter_dict['role'] = danceRole

            current_matches = 0
            overlap_matches = 0
            nonconcurrent_filter = {'event__endTime__lte': timezone.now()}

            if registration:

                if isinstance(registration,Registration):
                    current_matches = registration.eventregistration_set.filter(**filter_dict).count()
                elif isinstance(registration,TemporaryRegistration):
                    current_matches = registration.temporaryeventregistration_set.filter(**filter_dict).count()

                nonconcurrent_filter = {'event__endTime__lte': registration.firstSeriesStartTime}
                overlap_matches = cust_priors.filter(**filter_dict).exclude(**nonconcurrent_filter).filter(
                    event__startTime__lte=registration.lastSeriesEndTime,
                ).count()

            priors_matches = cust_priors.filter(**filter_dict).filter(**nonconcurrent_filter).count()

            # The number of matches depends on the concurrency rule for this item
            if item.concurrentRule == item.ConcurrencyRule.prohibited:
                matches = priors_matches
            elif item.concurrentRule == item.ConcurrencyRule.allowOneOverlapClass:
                matches = priors_matches + \
                    cust_priors.filter(**filter_dict).exclude(**nonconcurrent_filter).filter(
                        event__startTime__lte=registration.getTimeOfClassesRemaining(1)
                    ).count()
            elif item.concurrentRule == item.ConcurrencyRule.allowTwoOverlapClasses:
                matches = priors_matches + \
                    cust_priors.filter(**filter_dict).exclude(**nonconcurrent_filter).filter(
                        event__startTime__lte=registration.getTimeOfClassesRemaining(2)
                    ).count()
            elif item.concurrentRule == item.ConcurrencyRule.allowed:
                matches = priors_matches + overlap_matches + \
                    (current_matches if isinstance(registration,TemporaryRegistration) else 0)
            elif item.concurrentRule == item.ConcurrencyRule.required:
                matches = overlap_matches + current_matches

            if matches >= item.quantity:
                # If this is an 'or' or a 'not' requirement, then we are done
                if self.booleanRule == self.BooleanChoice.booleanOr:
                    return True
                if self.booleanRule == self.BooleanChoice.booleanNot:
                    return False
            else:
                # If this is an 'and' requirement and we didn't meet, then we are done
                if self.booleanRule == self.BooleanChoice.booleanAnd:
                    return False

        # If we got this far, then either all 'and' requirements were met, or all 'or' and 'not' requirements were not met
        if self.booleanRule == self.BooleanChoice.booleanOr or self.requirementitem_set.count() == 0:
            return False
        return True

    def clean(self):
        if self.applicableClass and self.applicableLevel:
            raise ValidationError(_('Requirement must be for a specific class or for a dance level; it cannot be for both.'))
        if not self.applicableClass and not self.applicableLevel:
            raise ValidationError(_('Requirement must apply to either a specific class or to a dance level.'))

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = _('Class requirement')
        verbose_name = _('Class requirements')
        permissions = (
            ('ignore_requirements',_('Can register users for series regardless of any prerequisites or requirements')),
        )


class RequirementItem(models.Model):
    ''' Each component of a requirement is one of these '''

    requirement = models.ForeignKey(Requirement,verbose_name=_('Requirement'),null=True)

    class ConcurrencyRule(DjangoChoices):
        prohibited = ChoiceItem('P',_('Must have previously taken'))
        allowOneOverlapClass = ChoiceItem('1',_('May register/begin with one class remaining'))
        allowTwoOverlapClasses = ChoiceItem('1',_('May register/begin with two classes remaining'))
        allowed = ChoiceItem('A',_('Concurrent registration allowed'))
        required = ChoiceItem('R',_('Concurrent registration required'))

    quantity = models.PositiveSmallIntegerField(_('Quantity'),default=1)
    requiredLevel = models.ForeignKey(DanceTypeLevel,null=True,blank=True,verbose_name=_('Required Dance type/level'))
    requiredClass = models.ForeignKey(ClassDescription,null=True,blank=True,verbose_name=_('Required class'))

    concurrentRule = models.CharField(_('Concurrency Rule'),max_length=1,choices=ConcurrencyRule.choices,default=ConcurrencyRule.prohibited)

    def clean(self):
        if self.requiredLevel and self.requiredClass:
            raise ValidationError(_('Requirement item cannot specify both a level and a class.'))
        if not self.requiredLevel and not self.requiredClass:
            raise ValidationError(_('Either a level or a class must be required.'))

    class Meta:
        verbose_name = _('Requirement item')
        verbose_name_plural = _('Requirement items')


class CustomerRequirement(models.Model):
    '''
    This class allows for override of requirements on a per-customer basis.
    '''
    customer = models.ForeignKey(Customer,verbose_name=_('Customer'))
    requirement = models.ForeignKey(Requirement,verbose_name=_('Requirement'))
    role = models.ForeignKey(DanceRole,null=True,blank=True,verbose_name=_('Dance role'),help_text=_('Role must be specified only for requirements for which roles are enforced.'))

    met = models.BooleanField(_('Meets Requirement'),default=True,help_text=_('If unchecked, then the customer explicitly does not meet the requirement, regardless of whether they meet its parameters.'))
    comments = models.TextField(_('Comments/Notes'),null=True,blank=True)

    submissionDate = models.DateTimeField(_('Submission date'),auto_now_add=True)
    modifiedDate = models.DateTimeField(_('Last modified date'),auto_now=True)

    def clean(self):
        if self.requirement.roleEnforced and not self.role:
            raise ValidationError(_('Since roles are enforced for this requirement, you must specify the customer\'s dance role.'))

    class Meta:
        unique_together = ('customer','requirement','role')
        verbose_name = _('Customer-level requirement record')
        verbose_name_plural = _('Customer-level requirement records')



from django.core.urlresolvers import reverse
from django.utils.translation import ugettext_lazy as _

from cms.toolbar_pool import toolbar_pool
from cms.toolbar_base import CMSToolbar
from cms.toolbar.items import Break


@toolbar_pool.register
class PrivateEventsToolbar(CMSToolbar):
    ''' Adds private events links to the Staff toolbar and category links to the related items toolbar '''

    def populate(self):
        events_menu = self.toolbar.get_or_create_menu('core-events',_('Events'))

        # Place the "School Calendar" link after the Registration Page link, before the break
        position = events_menu.find_first(Break,identifier='post_registration_break') or 0
        events_menu.add_link_item(_('School Calendar'), url=reverse('privateCalendar'), position=position)

        if self.request.user.has_perm('private_events.change_privateeventcategory'):
            related_menu = events_menu.get_or_create_menu('core-events-related',_('Related Items'))
            related_menu.add_link_item(_('Private Event Categories'),url=reverse('admin:private_events_privateeventcategory_changelist'))



from django.http import JsonResponse, HttpResponseBadRequest
from django.utils.translation import ugettext_lazy as _
from django.core.exceptions import ObjectDoesNotExist
from django.contrib.auth.models import User
from django.contrib.sites.shortcuts import get_current_site
from django.utils.functional import SimpleLazyObject
from django.utils import timezone

import logging
from paypalrestsdk import Payment
from paypalrestsdk.exceptions import ResourceNotFound
from datetime import timedelta

from danceschool.core.models import TemporaryRegistration, Invoice
from danceschool.core.constants import getConstant, INVOICE_VALIDATION_STR

from .models import PaypalPaymentRecord


# Define logger for this file
logger = logging.getLogger(__name__)


def createPaypalPayment(request):
    '''
    This view handles the creation of Paypal Express Checkout Payment objects.

    All Express Checkout payments must either be associated with a pre-existing Invoice
    or a registration, or they must have an amount and type passed in the post data
    (such as gift certificate payment requests).
    '''
    logger.info('Received request for Paypal Express Checkout payment.')

    invoice_id = request.POST.get('invoice_id')
    tr_id = request.POST.get('reg_id')
    amount = request.POST.get('amount')
    submissionUserId = request.POST.get('user_id')
    transactionType = request.POST.get('transaction_type')
    taxable = request.POST.get('taxable', False)

    # If a specific amount to pay has been passed, then allow payment
    # of that amount.
    if amount:
        try:
            amount = float(amount)
        except ValueError:
            logger.error('Invalid amount passed')
            return HttpResponseBadRequest()

    # Parse if a specific submission user is indicated
    submissionUser = None
    if submissionUserId:
        try:
            submissionUser = User.objects.get(id=int(submissionUserId))
        except (ValueError, ObjectDoesNotExist):
            logger.warning('Invalid user passed, submissionUser will not be recorded.')

    try:
        # Invoice transactions are usually payment on an existing invoice.
        if invoice_id:
            this_invoice = Invoice.objects.get(id=invoice_id)
            this_description = _('Invoice Payment: %s' % this_invoice.id)
            if not amount:
                amount = this_invoice.outstandingBalance
        # This is typical of payment at the time of registration
        elif tr_id:
            tr = TemporaryRegistration.objects.get(id=int(tr_id))
            tr.expirationDate = timezone.now() + timedelta(minutes=getConstant('registration__sessionExpiryMinutes'))
            tr.save()
            this_invoice = Invoice.get_or_create_from_registration(tr, submissionUser=submissionUser)
            this_description = _('Registration Payment: #%s' % tr_id)
            if not amount:
                amount = this_invoice.outstandingBalance
        # All other transactions require both a transaction type and an amount to be specified
        elif not transactionType or not amount:
            logger.error('Insufficient information passed to createPaypalPayment view.')
            raise ValueError
        else:
            # Gift certificates automatically get a nicer invoice description
            if transactionType == 'Gift Certificate':
                this_description = _('Gift Certificate Purchase')
            else:
                this_description = transactionType
            this_invoice = Invoice.create_from_item(
                float(amount),
                this_description,
                submissionUser=submissionUser,
                calculate_taxes=(taxable is not False),
                transactionType=transactionType,
            )
    except (ValueError, ObjectDoesNotExist) as e:
        logger.error('Invalid registration information passed to createPaypalPayment view: (%s, %s, %s)' % (invoice_id, tr_id, amount))
        logger.error(e)
        return HttpResponseBadRequest()

    this_currency = getConstant('general__currencyCode')

    this_total = min(this_invoice.outstandingBalance, amount)
    this_subtotal = this_total - this_invoice.taxes

    this_transaction = {
        'amount': {
            'total': round(this_total,2),
            'currency': this_currency,
            'details': {
                'subtotal': round(this_subtotal,2),
                'tax': round(this_invoice.taxes,2),
            },
        },
        'description': str(this_description),
        'item_list': {
            'items': []
        }
    }

    for item in this_invoice.invoiceitem_set.all():

        if not getConstant('registration__buyerPaysSalesTax'):
            this_item_price = item.grossTotal - item.taxes
        else:
            this_item_price = item.grossTotal

        this_transaction['item_list']['items'].append({
            'name': str(item.name),
            'price': round(this_item_price,2),
            'tax': round(item.taxes,2),
            'currency': this_currency,
            'quantity': 1,
        })

    # Because the Paypal API requires that the subtotal add up to the sum of the item
    # totals, we must add a negative line item for discounts applied, and a line item
    # for the remaining balance if there is to be one.
    if this_invoice.grossTotal != this_invoice.total:
        this_transaction['item_list']['items'].append({
            'name': str(_('Total Discounts')),
            'price': round(this_invoice.total,2) - round(this_invoice.grossTotal,2),
            'currency': this_currency,
            'quantity': 1,
        })
    if this_invoice.amountPaid > 0:
        this_transaction['item_list']['items'].append({
            'name': str(_('Previously Paid')),
            'price': -1 * round(this_invoice.amountPaid,2),
            'currency': this_currency,
            'quantity': 1,
        })
    if amount != this_invoice.outstandingBalance:
        this_transaction['item_list']['items'].append({
            'name': str(_('Remaining Balance After Payment')),
            'price': round(amount,2) - round(this_invoice.outstandingBalance,2),
            'currency': this_currency,
            'quantity': 1,
        })

    # Paypal requires the Payment request to include redirect URLs.  Since
    # the plugin can handle actual redirects, we just pass the base URL for
    # the current site.
    site = SimpleLazyObject(lambda: get_current_site(request))
    protocol = 'https' if request.is_secure() else 'http'
    base_url = SimpleLazyObject(lambda: "{0}://{1}".format(protocol, site.domain))

    payment = Payment({
        'intent': 'sale',
        'payer': {
            'payment_method': 'paypal'
        },
        'transactions': [this_transaction],
        'redirect_urls': {
            'return_url': str(base_url),
            'cancel_url': str(base_url),
        }
    })

    if payment.create():
        logger.info('Paypal payment object created.')

        if this_invoice:
            this_invoice.status = Invoice.PaymentStatus.authorized
            this_invoice.save()

            # We just keep a record of the ID and the status, because the
            # API can be used to look up everything else.
            PaypalPaymentRecord.objects.create(
                paymentId=payment.id,
                invoice=this_invoice,
                status=payment.state,
            )

        return JsonResponse(payment.to_dict())
    else:
        logger.error('Paypal payment object not created.')
        logger.error(payment)
        logger.error(payment.error)
        if this_invoice:
            this_invoice.status = Invoice.PaymentStatus.error
            this_invoice.save()
        return HttpResponseBadRequest()


def executePaypalPayment(request):
    paymentId = request.POST.get('paymentID')
    payerId = request.POST.get('payerID')
    addSessionInfo = request.POST.get('addSessionInfo',False)
    successUrl = request.POST.get('successUrl')

    try:
        payment_record = PaypalPaymentRecord.objects.get(paymentId=paymentId)
        payment = payment_record.getPayment()
        this_invoice = payment_record.invoice
    except (ResourceNotFound, ObjectDoesNotExist):
        logger.error('Unable to find local record of payment: %s' % paymentId)
        return HttpResponseBadRequest()

    if payment.execute({'payer_id': payerId}):
        payment_record.status = payment.state
        payment_record.payerId = payerId
        payment_record.save()

        this_invoice.processPayment(
            amount=float(payment.transactions[0].amount.total),
            fees=float(payment.transactions[0].related_resources[0].sale.transaction_fee.value),
            paidOnline=True,
            methodName='Paypal Express Checkout',
            methodTxn=paymentId,
            notify=payment.payer.payer_info.email,
        )

        if addSessionInfo:
            paymentSession = request.session.get(INVOICE_VALIDATION_STR, {})

            paymentSession.update({
                'invoiceID': str(this_invoice.id),
                'amount': float(payment.transactions[0].amount.total),
                'successUrl': successUrl,
            })
            request.session[INVOICE_VALIDATION_STR] = paymentSession

        return JsonResponse({'paid': True})
    else:
        this_invoice.status = Invoice.PaymentStatus.error
        this_invoice.save()
        payment_record.status = payment.state
        payment_record.save()
        logger.error()



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-06-14 16:07
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0005_auto_20170614_1607'),
        ('paypal', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='PaymentRecord',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('paymentId', models.CharField(max_length=50, unique=True, verbose_name='Paypal Payment ID')),
                ('payerId', models.CharField(blank=True, max_length=50, null=True, verbose_name='Paypal Payer ID')),
                ('status', models.CharField(blank=True, max_length=30, null=True, verbose_name='Current status')),
                ('creationDate', models.DateTimeField(auto_now_add=True, verbose_name='Created')),
                ('modifiedDate', models.DateTimeField(auto_now=True, verbose_name='Last updated')),
            ],
        ),
        migrations.RemoveField(
            model_name='ipncartitem',
            name='ipn',
        ),
        migrations.RemoveField(
            model_name='ipncartitem',
            name='revenueItem',
        ),
        migrations.RemoveField(
            model_name='ipnmessage',
            name='finalRegistration',
        ),
        migrations.RemoveField(
            model_name='ipnmessage',
            name='paypalInvoice',
        ),
        migrations.RemoveField(
            model_name='ipnmessage',
            name='priorTransaction',
        ),
        migrations.RemoveField(
            model_name='ipnmessage',
            name='registration',
        ),
        migrations.RemoveField(
            model_name='paynowformmodel',
            name='cancellationPage',
        ),
        migrations.DeleteModel(
            name='Invoice',
        ),
        migrations.DeleteModel(
            name='IPNCartItem',
        ),
        migrations.DeleteModel(
            name='IPNMessage',
        ),
        migrations.AddField(
            model_name='paymentrecord',
            name='invoice',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='paypalpayments', to='core.Invoice', verbose_name='Invoice'),
        ),
    ]



from django import forms
from django.contrib import messages
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.db.models import Q
from django.utils.encoding import force_text
from django.forms.widgets import CheckboxSelectMultiple, CheckboxInput, mark_safe, Select
from django.utils.html import conditional_escape, format_html
from django.utils.translation import ugettext_lazy as _, ugettext

from itertools import chain
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, Div, Field, HTML, Hidden, Submit
from dal import autocomplete
from random import random
import json
import logging
from djangocms_text_ckeditor.widgets import TextEditorWidget

from .models import EventStaffMember, SubstituteTeacher, Event, EventOccurrence, Series, SeriesTeacher, Instructor, EmailTemplate, Location, Customer, Invoice, get_defaultEmailName, get_defaultEmailFrom
from .constants import HOW_HEARD_CHOICES, getConstant, REG_VALIDATION_STR
from .signals import check_student_info
from .utils.emails import get_text_for_html
from .utils.timezone import ensure_localtime

# Define logger for this file
logger = logging.getLogger(__name__)


class LocationWithDataWidget(Select):
    '''
    Override render_option to permit extra data of default capacity
    and room options to be used by JQuery.
    '''

    def render_option(self, selected_choices, option_value, option_label):
        if option_value is None:
            option_value = ''
        option_value = force_text(option_value)
        if option_value in selected_choices:
            selected_html = mark_safe(' selected="selected"')
            if not self.allow_multiple_selected:
                # Only allow for a single selection.
                selected_choices.remove(option_value)
        else:
            selected_html = ''

        # Pass the default location capacity as an option
        if option_value:
            this_location = Location.objects.filter(id=int(option_value)).first()
            defaultCapacity = this_location.defaultCapacity
            room_options = [{'id': x.id, 'name': x.name, 'defaultCapacity': x.defaultCapacity} for x in this_location.room_set.all()]

            extra_value_data = format_html(
                ' data-defaultCapacity="{}" data-roomOptions="{}"',
                defaultCapacity, json.dumps(room_options))
        else:
            extra_value_data = ''

        return format_html('<option value="{}"{}{}>{}</option>',
                           option_value,
                           mark_safe(selected_html),
                           extra_value_data,
                           force_text(option_label))


class CheckboxSelectMultipleWithDisabled(CheckboxSelectMultiple):
    """
    Subclass of Django's checkbox select multiple widget that allows disabling checkbox-options,
    as well as marking some options as subject to override. To disable an option, pass a dict
    instead of a string for its label, of the form: {'label': 'option label', 'disabled': True}.
    To make an option part of a separate "override" choice set, add a dictionary key {'override': True}
    """

    def render(self, name, value, attrs=None, choices=()):
        if value is None:
            value = []
        has_id = attrs and 'id' in attrs
        final_attrs = self.build_attrs(attrs, extra_attrs={'name': name})
        output = [u'',]

        # Separate out regular choices and override-only choices
        all_choices = list(chain(self.choices, choices))

        override_choices = [
            x for x in all_choices if
            isinstance(x[1],dict) and
            'override' in x[1].keys() and
            x[1]['override']
        ]

        regular_choices = [
            x for x in all_choices if x not in override_choices
        ]

        # Normalize to strings
        str_values = set([force_text(v,encoding='utf-8') for v in value])

        if regular_choices:
            output.append(u'<ul class="list-unstyled">')

            for i, (option_value, option_label) in enumerate(regular_choices):
                if 'disabled' in final_attrs:
                    del final_attrs['disabled']
                if isinstance(option_label, dict):
                    if dict.get(option_label, 'disabled'):
                        final_attrs = dict(final_attrs,disabled='disabled')
                    option_label = option_label['label']
                # If an ID attribute was given, add a numeric index as a suffix,
                # so that the checkboxes don't all have the same ID attribute.
                if has_id:
                    final_attrs = dict(final_attrs, id='%s_%s' % (attrs['id'], i))
                    label_for = u' for="%s"' % final_attrs['id']
                else:
                    label_for = ''
                cb = CheckboxInput(final_attrs, check_test=lambda value: value in str_values)
                option_value = force_text(option_value,encoding='utf-8')
                rendered_cb = cb.render(name, option_value)
                option_label = conditional_escape(force_text(option_label,encoding='utf=8'))
                output.append(u'<li><label%s>%s %s</label></li>' % (label_for, rendered_cb, option_label))

            output.append(u'</ul>')

        if override_choices:
            # Determine whether or not to add a Submit button
            submit_button_flag = False

            # Create an ID for the collapse
            if has_id:
                collapse_id = 'override_' + str(attrs['id'])
            else:
                collapse_id = 'override_' + str(int(random() * 10.0**12))

            output.append(u'<button class="btn btn-outline-secondary btn-sm mb-4" type="button" data-toggle="collapse" data-target="#%(id)s">%(string)s</button><div class="collapse" id="%(id)s"><ul class="list-unstyled">' % {'id': collapse_id, 'string': _('Additional Choices')})

            for i, (option_value, option_label) in enumerate(override_choices):
                if 'disabled' in final_attrs:
                    del final_attrs['disabled']
                if isinstance(option_label, dict):
                    if dict.get(option_label, 'disabled'):
                        final_attrs = dict(final_attrs,disabled='disabled')
                    if dict.get(option_label, 'closed'):
                        submit_button_flag = True
                    option_label = option_label['label']
                # If an ID attribute was given, add a numeric index as a suffix,
                # so that the checkboxes don't all have the same ID attribute.
                if has_id:
                    final_attrs = dict(final_attrs, id='%s_%s' % (attrs['id'], i))
                    label_for = u' for="%s"' % final_attrs['id']
                else:
                    label_for = ''
                cb = CheckboxInput(final_attrs, check_test=lambda value: value in str_values)
                option_value = force_text(option_value,encoding='utf=8')
                rendered_cb = cb.render(name, option_value)
                option_label = conditional_escape(force_text(option_label,encoding='utf=8'))
                output.append(u'<li><label%s>%s %s</label></li>' % (label_for, rendered_cb, option_label))
            if submit_button_flag:
                output.append(u'<input class="btn btn-outline-primary btn-sm" type="submit" value="%s &raquo;" />' % _('Register now'))
            output.append(u'</ul></div>')

        return mark_safe(u'\n'.join(output))


class CheckboxSeriesChoiceField(forms.MultipleChoiceField):
    '''
    Inherits from ChoiceField, and uses the widget above, but also
    cleans to raise an error if a user registers as both a lead and a follow,
    or if they register for the whole class plus a drop-in.
    '''
    widget = CheckboxSelectMultipleWithDisabled


class ClassChoiceForm(forms.Form):
    '''
    This is the form that customers use to select classes.
    '''

    def __init__(self,*args,**kwargs):
        openEvents = kwargs.pop('openEvents',Event.objects.none())
        closedEvents = kwargs.pop('closedEvents', Event.objects.none())
        user = kwargs.pop('user',None)

        # Initialize a default (empty) form to fill
        super(ClassChoiceForm, self).__init__(*args, **kwargs)

        # Allow users with appropriate permissions to process door registrations.
        if user and user.has_perm('core.accept_door_payments'):
            self.fields['payAtDoor'] = forms.BooleanField(required=False,label=_('Door/Invoice Registration'))

        if user and user.has_perm('core.override_register_closed'):
            choice_set = openEvents | closedEvents
        else:
            choice_set = openEvents

        # Only the keys passed in this property will be entered into session data.
        # This prevents injection of unknown values into the registration process.
        self.permitted_event_keys = ['role',]

        for event in choice_set:
            field_choices = []

            # Get the set of roles for registration.  If custom roles and capacities
            # are provided, those will be used.  Or, if the DanceType of a Series
            # provides default roles, those will be used.  Otherwise, a single role will
            # be defined as 'Register' .
            eventRoles = event.eventrole_set.all()
            roles = []
            if eventRoles.count() > 0:
                roles = [x.role for x in eventRoles]
            elif isinstance(event,Series) and event.classDescription.danceTypeLevel.danceType.roles.count() > 0:
                roles = event.classDescription.danceTypeLevel.danceType.roles.all()

            # Add one choice per role
            for role in roles:
                this_label = {
                    'label': '%s (%s registered)' % (role.pluralName,event.numRegisteredForRole(role))
                }
                if event.soldOutForRole(role):
                    this_label = {'label': _('%s sold out!') % role.pluralName, 'disabled': True}
                    if user.has_perm('core.override_register_soldout'):
                        this_label['disabled'] = False
                        this_label['override'] = True
                if event in closedEvents:
                    this_label['closed'] = True
                    this_label['override'] = True
                field_choices.append(
                    (json.dumps({'role': role.id,}),this_label)
                )
            # If no choices, then add a general Register choice
            if not roles:
                this_label = {'label': _('Register (%s registered)') % event.numRegistered}
                if event.soldOut:
                    this_label = {'label': _('Sold out!'), 'disabled': True}
                    if user.has_perm('core.override_register_soldout'):
                        this_label['disabled'] = False
                        this_label['override'] = True
                if event in closedEvents:
                    this_label['closed'] = True
                    this_label['override'] = True
                field_choices.append(
                    (json.dumps({'role': None}),this_label)
                )

            # Add drop-in choices if they are available and if this user has permission.
            # If the user only has override permissions, add the override collapse only.
            # Note that this works because django-polymorphic returns the subclass.
            if isinstance(event,Series):
                if event.allowDropins and user and user.has_perm('core.register_dropins'):
                    for occurrence in event.eventoccurrence_set.all():
                        field_choices += ((json.dumps({'dropin_' + str(occurrence.id): True}), _('Drop-in: ') + ensure_localtime(occurrence.startTime).strftime('%B %-d')),)
                        self.permitted_event_keys.append('dropin_' + str(occurrence.id))
                elif (user and user.has_perm('core.override_register_dropins')):
                    for occurrence in event.eventoccurrence_set.all():
                        field_choices += ((json.dumps({'dropin_' + str(occurrence.id): True}),{'label': _('Drop-in: ') + ensure_localtime(occurrence.startTime).strftime('%B %-d'),'override':True}),)
                        self.permitted_event_keys.append('dropin_' + str(occurrence.id))

            self.fields['event_' + str(event.id)] = CheckboxSeriesChoiceField(
                label=event.name,
                choices=field_choices,
                required=False,
            )

    def clean(self):
        # Check that the registration is not empty
        cleaned_data = super(ClassChoiceForm,self).clean()
        hasContent = False

        for key,value in cleaned_data.items():
            if value and key != 'payAtDoor':
                hasContent = True
            if isinstance(value,list):
                # Ignore any passed value that is not a dictionary
                value_dict_list = [json.loads(x) for x in value if isinstance(json.loads(x),dict)]

                # Get the list of roles -- if more than one, then raise an error
                roles = [y.pop('role') for y in value_dict_list if y.get('role')]
                value_dict = {}
                for v in value_dict_list:
                    value_dict.update(v)

                # Get the list of dropIns
                dropIns = [k.replace('dropin_','') for k in value_dict.keys() if 'dropin_' in k]

                if len(roles) > 1:
                    raise ValidationError(_('Must select only one role.'),code='invalid')
                elif len(roles) == 1 and len(dropIns) > 0:
                    raise ValidationError(_('Cannot register for drop-in classes and also for the entire series.'),code='invalid')
        if not hasContent:
            raise ValidationError(_('Must register for at least one class or series.'))


class RegistrationContactForm(forms.Form):
    '''
    This is the form customers use to fill out their contact info.
    '''

    firstName = forms.CharField(label=_('First Name'))
    lastName = forms.CharField(label=_('Last Name'))
    email = forms.EmailField()
    phone = forms.CharField(required=False,label=_('Telephone (optional)'),help_text=_('We may use this to notify you in event of a cancellation.'))
    student = forms.BooleanField(required=False,label=_('I am a student'), help_text=_('Photo ID is required at the door'))
    mailList = forms.BooleanField(required=False,label=_('Add me to the mailing list'))
    agreeToPolicies = forms.BooleanField(required=True,label=_('<strong>I agree to all policies (required)</strong>'),help_text=_('By checking, you agree to abide by all policies.'))
    gift = forms.CharField(required=False,label=_('Voucher ID'))
    howHeardAboutUs = forms.ChoiceField(choices=HOW_HEARD_CHOICES,required=False,label=_('How did you hear about us?'),help_text=_('Optional'))
    comments = forms.CharField(widget=forms.Textarea,required=False,label=_('Comments'),help_text=_('Add anything else you\'d like to tell us.'))

    def get_top_layout(self):

        top_layout = Layout(
            Div(
                Field('firstName', wrapper_class='col'),
                Field('lastName', wrapper_class='col'),
                css_class='row'
            ),
            Div(
                Field('email', wrapper_class='col'),
                Field('phone',wrapper_class='col'),
                css_class='row'
            ),
        )
        return top_layout

    def get_mid_layout(self):
        mid_layout = Layout(
            Div('agreeToPolicies','student',css_class='card card-body bg-light my-2'),
        )
        return mid_layout

    def get_bottom_layout(self):
        bottom_layout = Layout(
            Div(
                Field('gift', wrapper_class='col'),
                Field('howHeardAboutUs', wrapper_class='col'),
                css_class='row mt-4'
            ),
            'comments',
        )
        return bottom_layout

    def __init__(self,*args,**kwargs):
        self._request = kwargs.pop('request',None)
        self._registration = kwargs.pop('registration',None)
        user = getattr(self._request,'user',None)
        session = getattr(self._request,'session',{}).get(REG_VALIDATION_STR,{})

        super(RegistrationContactForm,self).__init__(*args,**kwargs)
        self._session = session

        self.helper = FormHelper()
        self.helper.form_method = 'post'
        self.helper.form_tag = False  # Our template must explicitly include the <form tag>

        if user and hasattr(user,'customer') and user.customer and not session.get('payAtDoor',False):
            # Input existing info for users who are logged in and have signed up before
            self.fields['firstName'].initial = user.customer.first_name or user.first_name
            self.fields['lastName'].initial = user.customer.last_name or user.last_name
            self.fields['email'].initial = user.customer.email or user.email
            self.fields['phone'].initial = user.customer.phone

        self.helper.layout = Layout(
            self.get_top_layout(),
            self.get_mid_layout(),
            self.get_bottom_layout(),
            Submit('submit',_('Complete Registration'))
        )

        # If a voucher ID was passed (i.e. a referral code), then populate the form
        # and clear the passed session value
        session['gift'] = ''
        if session.get('voucher_id'):
            self.fields['gift'].initial = session.get('voucher_id')
            session['voucher_id'] = None

    def is_valid(self):
        '''
        For this form to be considered valid, there must be not only no errors, but also no messages on
        the request that need to be shown.
        '''

        valid = super(RegistrationContactForm,self).is_valid()
        msgs = messages.get_messages(self._request)

        # We only want validation messages to show up once, so pop messages that have already show up
        # before checking to see if any messages remain to be shown.
        prior_messages = self._session.pop('prior_messages',[])
        remaining_messages = []

        for m in msgs:
            m_dict = {'message': m.message, 'level': m.level, 'extra_tags': m.extra_tags}
            if m_dict not in prior_messages:
                remaining_messages.append(m_dict)

        if remaining_messages:
            self._session['prior_messages'] = remaining_messages
            self._request.session.modified = True
            return False
        return valid

    def clean(self):
        super(RegistrationContactForm,self).clean()
        first = self.cleaned_data.get('firstName')
        last = self.cleaned_data.get('lastName')
        email = self.cleaned_data.get('email')

        # Check that this customer is not already registered for any of the Events in the list
        customer = Customer.objects.filter(
            first_name=first,
            last_name=last,
            email=email).first()

        if customer:
            eventids = [x.event.id for x in self._registration.temporaryeventregistration_set.all()]
            already_registered_list = customer.getSeriesRegistered().filter(id__in=eventids)
        else:
            already_registered_list = []

        if already_registered_list:
            error_list = '\n'.join(['<li>%s</li>' % (x.name,) for x in already_registered_list])
            raise ValidationError(ugettext(mark_safe('You are already registered for:\n<ul>\n%s\n</ul>\nIf you are registering another person, please enter their name.' % error_list)))

        # Allow other handlers to add validation errors to the form.  Also, by passing the request, we allow
        # those handlers to add messages to the request, which (for this form) are treated like errors in that
        # they prevent the form from being considered valid.
        check_student_info.send(sender=RegistrationContactForm,instance=self,formData=self.cleaned_data,request=self._request,registration=self._registration)

        return self.cleaned_data


class DoorAmountForm(forms.Form):
    '''
    This is the form that staff users fill out to indicate that they received a cash
    payment.  Upon this being marked, the registration is processed.
    '''

    submissionUser = forms.ModelChoiceField(queryset=User.objects.filter(Q(staffmember__isnull=False) | Q(is_staff=True)),required=True)

    paid = forms.BooleanField(label=_('Payment Received'),required=False)
    receivedBy = forms.ModelChoiceField(
        queryset=User.objects.filter(Q(staffmember__isnull=False) | Q(is_staff=True)),
        label=_('Payment received by:'),
        required=False,
        widget=autocomplete.ModelSelect2(
            url='autocompleteUser',
            attrs={
                # This will set the input placeholder attribute:
                'data-placeholder': _('Enter a user name'),
                # This will set the yourlabs.Autocomplete.minimumCharacters
                # options, the naming conversion is handled by jQuery
                'data-autocomplete-minimum-characters': 2,
                'data-widget-maximum-values': 4,
                'class': 'modern-style',
            }
        )
    )
    amountPaid = forms.FloatField(label=_('Amount Paid'),required=False)

    invoiceSent = forms.BooleanField(label=_('Send Invoice'),required=False)
    cashPayerEmail = forms.EmailField(label=_('Payer Email Address'),required=False)
    invoicePayerEmail = forms.EmailField(label=_('Payer Email Address'),required=False)
    discountAmount = forms.FloatField(required=False)

    def __init__(self,*args,**kwargs):
        user = kwargs.pop('user',None)
        payerEmail = kwargs.pop('payerEmail',None)
        doorPortion = kwargs.pop('doorPortion', None)
        invoicePortion = kwargs.pop('invoicePortion', None)
        discountAmount = kwargs.pop('discountAmount', None)

        subUser = getattr(user,'id',None)

        self.helper = FormHelper()
        self.helper.form_method = 'post'
        self.helper.form_tag = False  # Our template must explicitly include the <form tag>

        if doorPortion:
            door_layout = Layout(
                HTML("""
                    <div class="card mt-4">
                        <h6 class="card-header" role="tab" id="door_headingOne">
                            """ + str(_('Cash Payment')) + """
                        </h6>
                        <div class="card-body">
                    """),
                'paid',
                'receivedBy',
                'amountPaid',
                'cashPayerEmail',
                Submit('submit','Submit'),
                HTML("""
                        </div>
                    </div>
                """),
            )
        else:
            door_layout = Layout(HTML(""))

        if invoicePortion:
            invoice_layout = Layout(
                HTML("""
                    <div class="card mt-4">
                        <h6 class="card-header" role="tab" id="door_headingTwo">
                                """ + str(_('Send Invoice')) + """
                        </h6>
                        <div class="card-body">
                    """),
                'invoiceSent',
                'invoicePayerEmail',
                Hidden('discountAmount', discountAmount),
                Submit('submit','Submit'),
                HTML("""
                        </div>
                    </div>
                """),
            )
        else:
            invoice_layout = Layout(HTML(""))

        self.helper.layout = Layout(
            HTML('<div id="door_accordion" role="tablist" aria-multiselectable="true">'),
            Hidden('submissionUser',subUser),
            door_layout,
            invoice_layout,
            HTML('</div>')
        )

        kwargs.update(initial={
            'cashPayerEmail': payerEmail,
            'invoicePayerEmail': payerEmail,
            'receivedBy': subUser,
        })

        super(DoorAmountForm,self).__init__(*args, **kwargs)

    def clean_submissionUser(self):
        paid = self.data.get('paid') or None
        invoiceSent = self.data.get('invoiceSent') or None

        user_id = self.data.get('submissionUser') or None

        if user_id:
            user = User.objects.get(id=user_id)
        if not user_id or not user:
            raise ValidationError(_('submissionUser not found.'))

        elif paid and not user.has_perm('core.accept_door_payments'):
            raise ValidationError(_('Invalid user submitted door payment.'))
        elif invoiceSent and not user.has_perm('core.send_invoices'):
            raise ValidationError(_('Invalid user submitted invoice.'))
        return user

    def clean(self):
        form_data = self.cleaned_data

        logger.debug('Form Data:\n%s' % form_data)

        paid = form_data.get('paid')
        invoiceSent = form_data.get('invoiceSent')

        if paid and invoiceSent:
            raise ValidationError(_('Must choose either cash payment or invoice submission, not both.'))

        if not paid and not invoiceSent:
            raise ValidationError(_('Must select either cash payment or invoice submission.'))

        # Check for required values here because the form can be used in two ways.
        if paid:
            if not form_data.get('submissionUser'):
                raise ValidationError(_('Submission user is required.'))
            if not form_data.get('receivedBy'):
                raise ValidationError(_('Must specify recipient of the money.'))
            if not form_data.get('amountPaid'):
                raise ValidationError(_('Must specify the amount of the payment.'))

        if invoiceSent:
            if not form_data.get('submissionUser'):
                raise ValidationError(_('Submission user is required.'))
            if not form_data.get('invoicePayerEmail'):
                raise ValidationError(_('Must specify the email address of the invoice recipient.'))

        return form_data


class RefundForm(forms.ModelForm):
    '''
    This is the form that is used to allocate refunds across series and events.  If the Paypal app is installed, then it
    will also be used to submit refund requests to Paypal.  Note that most cleaning validation happens in Javascript.
    '''
    class Meta:
        model = Invoice
        fields = []

    def __init__(self, *args, **kwargs):
        super(RefundForm, self).__init__(*args, **kwargs)

        this_invoice = kwargs.pop('instance',None)

        for item in this_invoice.invoiceitem_set.all():
            initial = False
            if item.finalEventRegistration:
                initial = item.finalEventRegistration.cancelled
            item_max = item.total + item.taxes if this_invoice.buyerPaysSalesTax else item.total

            self.fields["item_cancelled_%s" % item.id] = forms.BooleanField(
                label=_('Cancelled'),required=False,initial=initial)
            self.fields['item_refundamount_%s' % item.id] = forms.FloatField(
                label=_('Refund Amount'),required=False,initial=(-1) * item.adjustments, min_value=0, max_value=item_max)

        self.fields['comments'] = forms.CharField(
            label=_('Explanation/Comments (optional)'),required=False,
            help_text=_('This information will be added to the comments on the invoice associated with this refund.'),
            widget=forms.Textarea(attrs={'placeholder': _('Enter explanation/comments...'), 'class': 'form-control'}))

        self.fields['id'] = forms.ModelChoiceField(
            required=True,queryset=Invoice.objects.filter(id=this_invoice.id),widget=forms.HiddenInput(),initial=this_invoice.id)

        self.fields['initial_refund_amount'] = forms.FloatField(
            required=True,initial=(-1) * this_invoice.adjustments,min_value=0,max_value=this_invoice.amountPaid + this_invoice.refunds,widget=forms.HiddenInput())

        self.fields['total_refund_amount'] = forms.FloatField(
            required=True,initial=0,min_value=0,max_value=this_invoice.amountPaid + this_invoice.refunds,widget=forms.HiddenInput())

    def clean_total_refund_amount(self):
        '''
        The Javascript should ensure that the hidden input is updated, but double check it here.
        '''
        initial = self.cleaned_data.get('initial_refund_amount', 0)
        total = self.cleaned_data['total_refund_amount']
        summed_refunds = sum([v for k,v in self.cleaned_data.items() if k.startswith('item_refundamount_')])

        if not self.cleaned_data.get('id'):
            raise ValidationError('ID not in cleaned data')

        if summed_refunds != total:
            raise ValidationError(_('Passed value does not match sum of allocated refunds.'))
        elif summed_refunds > self.cleaned_data['id'].amountPaid + self.cleaned_data['id'].refunds:
            raise ValidationError(_('Total refunds allocated exceed revenue received.'))
        elif total < initial:
            raise ValidationError(_('Cannot reduce the total amount of the refund.'))
        return total


class EmailContactForm(forms.Form):

    EMAIL_SENDTOSET_CHOICES = (('series',_('All students in one or more series')),('month',_('All Students in a given month')))
    RICH_TEXT_CHOICES = (('plain',_('Plain text email')),('HTML',_('HTML rich text email')))

    sendToSet = forms.ChoiceField(label=_('This email is for:'),widget=forms.RadioSelect,choices=EMAIL_SENDTOSET_CHOICES,required=False,initial='series')

    template = forms.ModelChoiceField(label=_('(Optional) Select a template'),required=False,queryset=EmailTemplate.objects.none())

    richTextChoice = forms.ChoiceField(label=_('Send this email as'),widget=forms.RadioSelect,choices=RICH_TEXT_CHOICES,required=True,initial='plain')

    subject = forms.CharField(max_length=100)

    message = forms.CharField(widget=forms.Textarea,required=False)
    html_message = forms.CharField(widget=TextEditorWidget,required=False)

    from_name = forms.CharField(max_length=50,initial=get_defaultEmailName)
    from_address = forms.EmailField(max_length=100,initial=get_defaultEmailFrom)
    cc_myself = forms.BooleanField(label=_('CC Myself:'),initial=True,required=False)
    month = forms.ChoiceField(label=_('Email all students registered in month:'),initial='',required=False)
    series = forms.MultipleChoiceField(label=_('Email all students registered in a current/recent series:'),initial='',required=False)
    testemail = forms.BooleanField(label=_('Test email:'),help_text=_('Send a test email to myself only.'),initial=False,required=False)

    def __init__(self, *args, **kwargs):
        user = kwargs.pop('user',None)
        months = kwargs.pop('months',[])
        recentseries = kwargs.pop('recentseries',[])
        customers = kwargs.pop('customers',[])

        super(EmailContactForm, self).__init__(*args, **kwargs)

        if customers:
            self.fields['customers'] = forms.MultipleChoiceField(
                required=True,
                label=_('Selected customers'),
                widget=forms.CheckboxSelectMultiple(),
                choices=[(x.id,'%s <%s>' % (x.fullName, x.email)) for x in customers]
            )
            self.fields['customers'].initial = [x.id for x in customers]
            self.fields.pop('month',None)
            self.fields.pop('series',None)
            self.fields.pop('sendToSet',None)

            # Move the customer list to the top of the form
            self.fields.move_to_end('customers',last=False)

        else:
            self.fields['month'].choices = months
            self.fields['series'].choices = recentseries

        if user:
            self.fields['template'].queryset = EmailTemplate.objects.filter(Q(groupRequired__isnull=True) | Q(groupRequired__in=user.groups.all())).filter(hideFromForm=False)

    def clean(self):
        # Custom cleaning ensures email is only sent to one of
        # a series, a month, or a set of customers
        super(EmailContactForm, self).clean()

        sendToSet = self.cleaned_data.get('sendToSet')
        customers = self.cleaned_data.get('customers')

        # We set to None and don't pop the keys out to prevent
        # KeyError issues with the subsequent view
        if sendToSet == 'series' or customers:
            self.cleaned_data['month'] = None
        if sendToSet == 'month' or customers:
            self.cleaned_data['series'] = None

        # If this is an HTML email, then ignore the plain text content
        # and replace it with plain text generated from the HTML.
        # If this is a plain text email, then ignore the HTML content.
        if self.cleaned_data['richTextChoice'] == 'HTML':
            if not self.cleaned_data['html_message']:
                raise ValidationError(_('Message is required.'))
            self.cleaned_data['message'] = get_text_for_html(self.cleaned_data['html_message'])
        if self.cleaned_data['richTextChoice'] == 'plain':
            if not self.cleaned_data['message']:
                raise ValidationError(_('Message is required.'))
            self.cleaned_data['html_message'] = None

        return self.cleaned_data

    class Media:
        js = ('js/emailcontact_sendToSet.js','js/emailcontact_ajax.js')


class SeriesTeacherChoiceField(forms.ModelChoiceField):
    '''
    This exists so that the validators for substitute teaching are not
    thrown off by the fact that the initial query is blank.
    '''

    def to_python(self,value):
        try:
            value = super(SeriesTeacherChoiceField,self).to_python(value)
        except (ValueError, ValidationError):
            key = self.to_field_name or 'pk'
            value = SeriesTeacher.objects.filter(**{key: value})
            if not value.exists():
                raise ValidationError(self.error_messages['invalid_choice'], code='invalid_choice')
            else:
                value = value.first()
        return value


class SeriesClassesChoiceField(forms.ModelMultipleChoiceField):
    '''
    This exists so that the validators for substitute teaching are not
    thrown off by the fact that the initial query is blank.
    '''

    def to_python(self, value):
        if not value:
            return []
        return list(self._check_values(value))

    def _check_values(self, value):
        """
        Given a list of possible PK values, returns a QuerySet of the
        corresponding objects. Raises a ValidationError if a given value is
        invalid (not a valid PK, not in the queryset, etc.)
        """
        key = self.to_field_name or 'pk'
        # deduplicate given values to avoid creating many querysets or
        # requiring the database backend deduplicate efficiently.
        try:
            value = frozenset(value)
        except TypeError:
            # list of lists isn't hashable, for example
            raise ValidationError(
                self.error_messages['list'],
                code='list',
            )
        for pk in value:
            try:
                self.queryset.filter(**{key: pk})
            except (ValueError, TypeError):
                raise ValidationError(
                    self.error_messages['invalid_pk_value'],
                    code='invalid_pk_value',
                    params={'pk': pk},
                )
        qs = EventOccurrence.objects.filter(**{'%s__in' % key: value})
        pks = set(force_text(getattr(o, key)) for o in qs)
        for val in value:
            if force_text(val) not in pks:
                raise ValidationError(
                    self.error_messages['invalid_choice'],
                    code='invalid_choice',
                    params={'value': val},
                )
        return qs


class SubstituteReportingForm(forms.ModelForm):

    def __init__(self, *args, **kwargs):
        user = kwargs.pop('user', None)

        # Ensure 'initial' is initialized to avoid KeyError issues and fill in the Sub category
        kwargs['initial'] = kwargs.get('initial',{})
        kwargs['initial'].update({'category': getConstant('general__eventStaffCategorySubstitute').id})

        # If the user is a staffMember, then populate the form with their info
        if hasattr(user,'staffmember'):
            kwargs['initial'].update({
                'staffMember': user.staffmember,
                'submissionUser': user,
            })

        super(SubstituteReportingForm,self).__init__(*args,**kwargs)
        self.fields['event'] = forms.ModelChoiceField(queryset=Series.objects.order_by('-startTime'))
        self.fields['staffMember'] = forms.ModelChoiceField(queryset=Instructor.objects.exclude(
            status__in=[
                Instructor.InstructorStatus.hidden,
                Instructor.InstructorStatus.retired,
                Instructor.InstructorStatus.retiredGuest
            ]).order_by('status','lastName','firstName'))
        self.fields['replacedStaffMember'] = SeriesTeacherChoiceField(queryset=SeriesTeacher.objects.none())
        self.fields['occurrences'] = SeriesClassesChoiceField(queryset=EventOccurrence.objects.none())
        self.fields['submissionUser'].widget = forms.HiddenInput()
        self.fields['category'].widget = forms.HiddenInput()

    def clean(self):
        '''
        This code prevents multiple individuals from substituting for the
        same class and class teacher.  It also prevents an individual from
        substituting for a class in which they are a teacher.
        '''
        super(SubstituteReportingForm,self).clean()

        occurrences = self.cleaned_data.get('occurrences',[])
        staffMember = self.cleaned_data.get('staffMember')
        replacementFor = self.cleaned_data.get('replacedStaffMember',[])
        event = self.cleaned_data.get('event')

        for occ in occurrences:
            for this_sub in occ.eventstaffmember_set.all():
                if this_sub.replacedStaffMember == replacementFor:
                    self.add_error('occurrences',ValidationError(_('One or more classes you have selected already has a substitute teacher for that class.'),code='invalid'))

        if event and staffMember:
            if staffMember in [x.staffMember for x in event.eventstaffmember_set.filter(category__in=[getConstant('general__eventStaffCategoryAssistant'),getConstant('general__eventStaffCategoryInstructor')])]:
                self.add_error('event',ValidationError(_('You cannot substitute teach for a class in which you were an instructor.'),code='invalid'))

    def validate_unique(self):
        '''
        We don't need to check the unique_together constraint in this form, because if the
        constraint is not satisfied, then the form will just update the existing instance
        in the save() method below.
        '''
        pass

    def save(self, commit=True):
        '''
        If a staff member is reporting substitute teaching for a second time, then we should update
        the list of occurrences for which they are a substitute on their existing EventStaffMember
        record, rather than creating a new record and creating database issues.
        '''
        existing_record = EventStaffMember.objects.filter(
            staffMember=self.cleaned_data.get('staffMember'),
            event=self.cleaned_data.get('event'),
            category=getConstant('general__eventStaffCategorySubstitute'),
            replacedStaffMember=self.cleaned_data.get('replacedStaffMember'),
        )
        if existing_record.exists():
            record = existing_record.first()
            for x in self.cleaned_data.get('occurrences'):
                record.occurrences.add(x)
            record.save()
            return record
        else:
            return super(SubstituteReportingForm,self).save()

    class Meta:
        model = SubstituteTeacher
        exclude = []

    class Media:
        js = ('js/substituteteacher_ajax.js',)


class InstructorBioChangeForm(forms.ModelForm):

    class Meta:
        model = Instructor
        fields = ['publicEmail','privateEmail','phone','availableForPrivates']


class RepeatEventForm(forms.Form):

    startDate = forms.DateField(label=_('First event occurs on'))
    repeatEvery = forms.IntegerField(label=_('Repeat every'),min_value=1, initial=1)
    periodicity = forms.ChoiceField(label=_('Period'),choices=(('D',_('Days')),('W',_('Weeks')),('M',_('Months')),), initial='W')
    quantity = forms.IntegerField(label=_('Repeat this many times'),min_value=1,max_value=99,required=False)
    endDate = forms.DateField(label=_('Repeat until this date'),required=False)

    def clean(self):
        startDate = self.cleaned_data.get('startDate')
        endDate = self.cleaned_data.get('endDate')
        quantity = self.cleaned_data.get('quantity')

        if endDate and not endDate >= startDate:
            self.add_error('endDate',ValidationError(_('End date must be after start date.')))

        if quantity and endDate:
            self.add_error('quantity',ValidationError(_('Please specify either a number of repeats or an end date, not both.')))


class InvoiceNotificationForm(forms.Form):
    '''
    This form just allows customers to deselect invoices for notification.
    '''

    def __init__(self,*args,**kwargs):
        invoices = kwargs.pop('invoices',Invoice.objects.none())

        # Initialize a default (empty) form to fill
        super(InvoiceNotificationForm, self).__init__(*args, **kwargs)

        for invoice in invoices:
            self.fields['invoice_%s' % invoice.id] = forms.BooleanField(label=invoice.id, required=False,initial=True)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-04-25 00:10
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('core', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='CustomerRequirement',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('met', models.BooleanField(default=True, help_text='If unchecked, then the customer explicitly does not meet the requirement, regardless of whether they meet its parameters.', verbose_name='Meets Requirement')),
                ('comments', models.TextField(blank=True, null=True, verbose_name='Comments/Notes')),
                ('submissionDate', models.DateTimeField(auto_now_add=True)),
                ('modifiedDate', models.DateTimeField(auto_now=True)),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Customer')),
            ],
        ),
        migrations.CreateModel(
            name='Requirement',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text="If a customer does not meet the requirement for a series, then this description will be used to explain the issue (e.g. 'Must have taken Lindy 1 to take Lindy 2').", max_length=300, verbose_name='Requirement name/description')),
                ('booleanRule', models.CharField(choices=[('&', 'Must meet all requirement items'), ('|', 'Must meet one or more requirement items'), ('!', 'Must not meet any requirement items')], default='&', help_text="If you select an option other than 'Must meet all requirements', then you can still enforce multiple requirements by adding another Requirement item.", max_length=3, verbose_name='How many items of this requirement must be met')),
                ('enforcementMethod', models.CharField(choices=[('N', 'Enforcement disabled'), ('W', 'Allow registration with warning'), ('E', 'Raise error and do not allow registration')], default='W', max_length=1, verbose_name='Enforcement method')),
                ('roleEnforced', models.BooleanField(default=False, help_text='If checked, then in order to meet the requirement, the customer must have met all individual requirements in the same dance role, and that must be the dance role for which they are registering.', verbose_name='Same dance role enforced')),
                ('submissionDate', models.DateTimeField(auto_now_add=True, verbose_name='Submission date')),
                ('modifiedDate', models.DateTimeField(auto_now=True, verbose_name='Last modified date')),
                ('applicableClass', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='core.ClassDescription', verbose_name='Applies to class')),
                ('applicableLevel', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='core.DanceTypeLevel', verbose_name='Applies to dance type/level')),
                ('applicableRole', models.ForeignKey(blank=True, help_text='Most requirements apply identically to all dance roles.  If this requirement does not, then specify the role to which it applies here.', null=True, on_delete=django.db.models.deletion.CASCADE, to='core.DanceRole', verbose_name='Applies only to role (optional)')),
            ],
            options={
                'permissions': (('ignore_requirements', 'Can register users for series regardless of any prerequisites or requirements'),),
            },
        ),
        migrations.CreateModel(
            name='RequirementItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('quantity', models.PositiveSmallIntegerField(default=1, verbose_name='Quantity')),
                ('concurrentRule', models.CharField(choices=[('P', 'Must have previously taken'), ('A', 'Concurrent registration allowed'), ('R', 'Concurrent registration required')], default='P', max_length=1, verbose_name='Concurrency Rule')),
                ('requiredClass', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='core.ClassDescription')),
                ('requiredLevel', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='core.DanceTypeLevel')),
                ('requirement', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='prerequisites.Requirement')),
            ],
        ),
        migrations.AddField(
            model_name='customerrequirement',
            name='requirement',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='prerequisites.Requirement'),
        ),
        migrations.AddField(
            model_name='customerrequirement',
            name='role',
            field=models.ForeignKey(blank=True, help_text='Role must be specified only for requirements for which roles are enforced.', null=True, on_delete=django.db.models.deletion.CASCADE, to='core.DanceRole'),
        ),
        migrations.AlterUniqueTogether(
            name='customerrequirement',
            unique_together=set([('customer', 'requirement', 'role')]),
        ),
    ]



from django.test import TestCase
from django.contrib.auth.models import User
from django.utils import timezone

from datetime import timedelta
from dynamic_preferences.registries import global_preferences_registry

from danceschool.core.models import DanceRole, DanceType, DanceTypeLevel, ClassDescription, PricingTier, Location, Instructor, Event, Series, EventStaffMember, EventOccurrence
from danceschool.core.constants import getConstant


class DefaultSchoolTestCase(TestCase):
    '''
    This class just sets up standard data for the school, and it can be
    inherited from, since many test classes in different apps may want to
    use this same test data.
    '''

    @classmethod
    def setUpTestData(cls):

        # Ensure that necessary constants are initialized and that all
        # needed categories are created within the database
        gp = global_preferences_registry.manager()
        gp.load_from_db()

        # Create Lead and Follow roles
        DanceRole.objects.create(name='Lead',order=1)
        DanceRole.objects.create(name='Follow',order=2)
        cls.defaultDanceRoles = DanceRole.objects.filter(name__in=['Lead','Follow'])
        cls.defaultDanceType = DanceType.objects.create(name='Lindy Hop',order=1)
        cls.defaultDanceType.roles = cls.defaultDanceRoles
        cls.defaultDanceType.save()

        # Create two levels for tests that involve different levels
        cls.levelOne = DanceTypeLevel.objects.create(name='Level 1', order=1, danceType=cls.defaultDanceType)
        cls.levelTwo = DanceTypeLevel.objects.create(name='Level 2', order=2, danceType=cls.defaultDanceType)

        # Create two ClassDescriptions for classes, one in each level
        cls.levelOneClassDescription = ClassDescription.objects.create(
            title='Test Level One Class',
            description='This is a test description',
            danceTypeLevel=cls.levelOne,
            slug='test-level-one',
        )
        cls.levelTwoClassDescription = ClassDescription.objects.create(
            title='Test Level Two Class',
            description='This is a test description',
            danceTypeLevel=cls.levelTwo,
            slug='test-level-two',
        )

        # Create a default PricingTier and a default Location
        cls.defaultPricing = PricingTier.objects.create(
            name='Default Pricing',
            onlinePrice=50,
            doorPrice=60,
            dropinPrice=10,
        )
        cls.defaultLocation = Location.objects.create(
            name='Default Location',
            status=Location.StatusChoices.active,
            address='This is a street address',
            city='Boston',
            state='MA',
            zip='02114',
            directions='These are directions to the default location.',
            defaultCapacity=50,
        )

        # Create a superuser and a non-staff user
        cls.superuser = User.objects.create_superuser(
            'admin',
            'admin@test.com',
            'pass',
            first_name='Frankie',
            last_name='Manning',
        )

        cls.nonStaffUser = User.objects.create_user(
            'regularuser',
            'user@domain.com',
            'pass',
            is_staff=False,
            first_name='New',
            last_name='Student',
        )

        # Make the superuser an Instructor
        cls.defaultInstructor = Instructor.objects.create(
            status=Instructor.InstructorStatus.roster,
            firstName='Frankie',
            lastName='Manning',
            userAccount=cls.superuser,
            publicEmail='admin@test.com',
            privateEmail='admin@test.com',
            bio='This is Frankie Manning.',
        )

    def create_series(self,**kwargs):
        """
        This method just creates a new series with the loaded class
        description that can be modified or used for various tests.
        """

        # Create one or more occurrences.  By default, the series
        # starts tomorrow, is a Level One Lindy Hop class with default
        # pricing and location, is enabled for registration, and is taught
        # by Frankie Manning.
        occurrences = max(kwargs.get('occurrences',1),1)
        startTime = kwargs.get('startTime', timezone.now() + timedelta(days=1))
        classDescription = kwargs.get('classDescription', self.levelOneClassDescription)
        pricingTier = kwargs.get('pricingTier', self.defaultPricing)
        location = kwargs.get('location', self.defaultLocation)
        status = kwargs.get('status', Event.RegStatus.enabled)
        instructors = kwargs.get('instructors', [self.defaultInstructor,])

        s = Series(
            classDescription=classDescription,
            pricingTier=pricingTier,
            location=location,
            status=status,
        )
        s.save()
        # Add an occurrence at the start Time
        # and if requested to set more than one occurrence, then
        # each additional occurrence is the day after the last one.
        for k in range(1,occurrences + 1):
            EventOccurrence.objects.create(
                event=s,
                startTime=startTime + timedelta(days=k - 1),
                endTime=startTime + timedelta(days=k - 1,hours=1)
            )
        # Add instructors (Frankie Manning by default)
        for i in instructors:
            seriesteacher = EventStaffMember.objects.create(
                event=s,
                category=getConstant('general__eventStaffCategoryInstructor'),
                staffMember=i,
            )
            seriesteacher.occurrences = s.eventoccurrence_set.all()
            seriesteacher.save()
        # Must save after adding event occurrences to ensure that
        # registration status is updated properly.
        s.save()
        return s

    def create_instructor(self,**kwargs):
        '''
        This method creates a new instructor (other than the default)
        for testing things like substitute teaching.
        '''
        status = kwargs.get('status', Instructor.InstructorStatus.roster)
        firstName = kwargs.get('firstName','Norma')
        lastName = kwargs.get('lastName','Miller')
        publicEmail = kwargs.get('publicEmail','norma@miller.com')
        privateEmail = kwargs.get('privateEmail', 'norma@miller.com')
        bio = kwargs.get('bio', 'This is Norma Miller.')
        userAccount = kwargs.get('userAccount', None)

        return Instructor.objects.create(
            status=status,
            firstName=firstName,
            lastName=lastName,
            userAccount=userAccount,
            publicEmail=publicEmail,
            privateEmail=privateEmail,
            bio=bio,
        )



from django.contrib import admin
from django.forms import ModelForm, ModelChoiceField, TextInput
from django.utils.translation import ugettext_lazy as _

from dal import autocomplete

from danceschool.core.models import Customer, Registration, TemporaryRegistration
from danceschool.core.admin import CustomerAdmin

from .models import VoucherCategory, Voucher, DanceTypeVoucher, ClassVoucher, CustomerGroupVoucher, CustomerVoucher, VoucherUse, TemporaryVoucherUse, VoucherCredit


class CustomerVoucherInlineForm(ModelForm):
    customer = ModelChoiceField(
        queryset=Customer.objects.all(),
        widget=autocomplete.ModelSelect2(
            url='autocompleteCustomer',
            attrs={
                # This will set the input placeholder attribute:
                'data-placeholder': _('Enter a customer name'),
                # This will set the yourlabs.Autocomplete.minimumCharacters
                # options, the naming conversion is handled by jQuery
                'data-minimum-input-length': 2,
                'data-max-results': 4,
                'class': 'modern-style',
            }
        )
    )

    class Meta:
        model = CustomerVoucher
        exclude = []


class CustomerGroupVoucherInline(admin.StackedInline):
    model = CustomerGroupVoucher
    extra = 1
    classes = ['collapse',]


class CustomerVoucherInline(admin.StackedInline):
    model = CustomerVoucher
    form = CustomerVoucherInlineForm
    extra = 1
    classes = ['collapse',]


class CustomerAdminWithVouchers(CustomerAdmin):
    inlines = CustomerAdmin.inlines + [CustomerVoucherInline]


class RegistrationVoucherInline(admin.TabularInline):
    model = VoucherUse
    extra = 0
    readonly_fields = ['voucher','amount']

    # Prevents adding new voucher uses without going through
    # the standard registration process
    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class TemporaryRegistrationVoucherInline(admin.TabularInline):
    model = TemporaryVoucherUse
    extra = 0
    readonly_fields = ['voucher','amount']

    # Prevents adding new voucher uses without going through
    # the standard registration process
    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class DanceTypeVoucherInline(admin.StackedInline):
    model = DanceTypeVoucher
    extra = 1
    classes = ['collapse',]


class ClassVoucherInline(admin.StackedInline):
    model = ClassVoucher
    extra = 1
    classes = ['collapse',]


class VoucherUseInline(admin.TabularInline):
    model = VoucherUse
    extra = 0
    readonly_fields = ['registration','amount']

    # Prevents adding new voucher uses without going through
    # the standard registration process.
    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class VoucherCreditInlineForm(ModelForm):

    class Meta:
        widgets = {
            'description': TextInput,
        }


class VoucherCreditInline(admin.TabularInline):
    model = VoucherCredit
    extra = 1
    fields = ['amount', 'description','creationDate']
    readonly_fields = ['creationDate',]

    form = VoucherCreditInlineForm


class VoucherAdmin(admin.ModelAdmin):
    inlines = [DanceTypeVoucherInline,ClassVoucherInline,CustomerGroupVoucherInline,CustomerVoucherInline,VoucherUseInline,VoucherCreditInline]
    list_display = ['voucherId','name','category','amountLeft','maxAmountPerUse','expirationDate','isEnabled','restrictions']
    list_filter = ['category','expirationDate','disabled','forFirstTimeCustomersOnly','forPreviousCustomersOnly']
    search_fields = ['voucherId','name','description']
    readonly_fields = ['refundAmount','creationDate']
    actions = ['enableVoucher','disableVoucher']

    fieldsets = (
        (None, {
            'fields': (('voucherId','category'),'name','description',('originalAmount','maxAmountPerUse'),),
        }),
        (_('Voucher Restrictions'), {
            'fields': ('expirationDate',('singleUse','forFirstTimeCustomersOnly','forPreviousCustomersOnly','disabled')),
        }),
        (_('Other Info'), {
            'classes': ('collapse',),
            'fields': ('creationDate','refundAmount'),
        }),
    )

    def isEnabled(self,obj):
        return obj.disabled is False
    isEnabled.short_description = _('Enabled')
    isEnabled.boolean = True

    def restrictions(self,obj):
        text = []
        if obj.singleUse:
            text.append(_('Single use'))
        if obj.forFirstTimeCustomersOnly:
            text.append(_('First-time customer'))
        if obj.forPreviousCustomersOnly:
            text.append(_('Previous customer'))
        if obj.customervoucher_set.all().exists():
            text.append(_('Specific customer'))
        if obj.classvoucher_set.all().exists():
            text.append(_('Specific class'))
        if obj.dancetypevoucher_set.all().exists():
            text.append(_('Specific dance type/level'))
        return ', '.join([str(x) for x in text])
    restrictions.short_description = _('Restrictions')

    def disableVoucher(self, request, queryset):
        rows_updated = queryset.update(disabled=True)
        if rows_updated == 1:
            message_bit = "1 voucher was"
        else:
            message_bit = "%s vouchers were" % rows_updated
        self.message_user(request, "%s successfully disabled." % message_bit)
    disableVoucher.short_description = _('Disable selected Vouchers')

    def enableVoucher(self, request, queryset):
        rows_updated = queryset.update(disabled=False)
        if rows_updated == 1:
            message_bit = "1 voucher was"
        else:
            message_bit = "%s vouchers were" % rows_updated
        self.message_user(request, "%s successfully enabled." % message_bit)
    enableVoucher.short_description = _('Enable selected Vouchers')


# This adds inlines to Registration and TemporaryRegistration without subclassing
admin.site._registry[Registration].inlines.insert(0,RegistrationVoucherInline)
admin.site._registry[TemporaryRegistration].inlines.insert(0,TemporaryRegistrationVoucherInline)

admin.site.register(VoucherCategory)
admin.site.register(Voucher,VoucherAdmin)
admin.site.unregister(Customer)
admin.site.register(Customer,CustomerAdminWithVouchers)
